use std::sync::Arc;
use std::{net::SocketAddr, path::PathBuf};

use clap::Parser;
use confique::{Config, Partial};
use eyre::{bail, Context};
use idcoop::cli::{handle_user_command, UserCommand};
use idcoop::config::{SecretConfig, SeparateSecretConfiguration};
use idcoop::store::IdCoopStore;
use idcoop::{config::Configuration, web};
use tracing_subscriber::fmt::format::FmtSpan;
use tracing_subscriber::layer::SubscriberExt;
use tracing_subscriber::util::SubscriberInitExt;

#[derive(Parser)]
#[command(version = env!("CARGO_PKG_VERSION"), author = env!("CARGO_PKG_AUTHORS"), about = env!("CARGO_PKG_DESCRIPTION"))]
struct Options {
    /// Path to a config file in TOML format. Can be specified multiple times. Optional (all options can also be specified in env vars).
    #[clap(short = 'c', long = "config", env = "IDCOOP_CONFIG")]
    config_files: Vec<PathBuf>,

    #[clap(short = 'S', long = "secrets", env = "IDCOOP_SECRETS")]
    secret_files: Vec<PathBuf>,

    #[clap(subcommand)]
    subcommand: Subcommand,
}

#[derive(Clone, Parser)]
enum Subcommand {
    /// Serves the idCoop service over HTTP.
    Serve {
        /// Manually choose an IP address and port to listen on. Overrides the configuration.
        #[clap(long = "bind")]
        bind: Option<SocketAddr>,
    },

    /// Manage users.
    User {
        #[clap(subcommand)]
        cmd: UserCommand,
    },
}

fn load_config_files<C: Config>(files: &[PathBuf]) -> eyre::Result<C> {
    let mut partial: C::Partial = Partial::from_env().context("failed to load env vars")?;

    for file in files {
        let cfg_file = confique::File::new(file)
            .with_context(|| format!("bad file path: {file:?}"))?
            .required()
            .load()
            .with_context(|| format!("failed to load file: {file:?}"))?;
        partial = partial.with_fallback(cfg_file);
    }

    partial = partial.with_fallback(Partial::default_values());

    C::from_partial(partial).context("failed to load config")
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> eyre::Result<()> {
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env()
                .unwrap_or_else(|_| "idcoop=debug,info".into()),
        )
        .with(tracing_subscriber::fmt::layer().with_span_events(FmtSpan::CLOSE))
        .init();

    let options = Options::parse();

    let mut config = load_config_files::<Configuration>(&options.config_files)
        .context("failed to load main config file")?;
    let mut sep_secrets = load_config_files::<SeparateSecretConfiguration>(&options.secret_files)
        .context("failed to load separated secrets file")?;

    // Load secrets into the OIDC Client Configurations
    for (oidc_client_id, oidc_client) in &mut config.oidc.clients {
        let specified_in_main = oidc_client.secret.is_some();
        match sep_secrets.oidc_secrets.remove(oidc_client_id) {
            Some(secret) => {
                if specified_in_main {
                    bail!("Client secret for OIDC Client {oidc_client_id:?} is specified both in main config file and in separated secret file!");
                }
                oidc_client.secret = Some(secret);
            }
            None => {
                if !specified_in_main {
                    bail!("Client secret for OIDC Client {oidc_client_id:?} is not specified!");
                }
            }
        }
    }

    match options.subcommand {
        Subcommand::Serve { bind } => {
            let Some(bind) = bind.or(config.listen.bind) else {
                bail!("neither --bind nor listen.bind config option set. Please specify a bind address!")
            };
            let store = IdCoopStore::connect(&config.postgres.connect)
                .await
                .context("Failed to connect to Postgres")?;
            let secrets = SecretConfig::try_new(&config).await?;
            web::serve(bind, Arc::new(store), Arc::new(config), Arc::new(secrets)).await?;
        }
        Subcommand::User { cmd } => {
            let store = IdCoopStore::connect(&config.postgres.connect)
                .await
                .context("Failed to connect to Postgres")?;
            handle_user_command(cmd, &config, &store).await?;
        }
    }

    Ok(())
}
