//! # idCoop Configuration
//!
//! [`confique`] is used to allow for a layered configuration.
//!
//! We have two files: a configuration file and a 'secrets' file.
//! The secrets file only contains secrets and is suitable for keeping encrypted and
//! separate from the main configuration in deployment tools.

use std::{
    collections::{BTreeMap, BTreeSet},
    net::SocketAddr,
    num::NonZeroU32,
    path::PathBuf,
};

use axum_client_ip::SecureClientIpSource;
use confique::Config;
use eyre::Context;
use governor::Quota;
use josekit::jwk::alg::rsa::RsaKeyPair;
use serde::{de::Error, Deserialize};

/// Root of the main configuration file.
///
/// This is split into several sections.
#[derive(Config)]
pub struct Configuration {
    /// Section of configuration controlling how the web interface listens.
    #[config(nested)]
    pub listen: ListenConfiguration,

    /// Section of configuration controlling how we connect to Postgres.
    #[config(nested)]
    pub postgres: PostgresConfiguration,

    /// Section of configuration for setting up OpenID Connect clients.
    #[config(nested)]
    pub oidc: OidcConfiguration,

    /// Section of configuration for configuring how passwords are hashed.
    #[config(nested)]
    pub password_hashing: PasswordHashingConfig,

    /// Section of configuration for configuring how requests are rate limited
    /// to prevent overload of the system or exceeding system requirements.
    #[config(nested)]
    pub ratelimits: RatelimitsConfig,
}

/// Configuration of values and clients for OpenID Connect.
#[derive(Config)]
pub struct OidcConfiguration {
    /// Our identity as an OAuth 2.1 / OIDC issuer.
    /// Example: `https://id.example.com`.
    /// Should really be unique.
    pub issuer: String,

    /// Path to an RSA keypair in PEM format.
    pub rsa_keypair: PathBuf,

    /// Map from client_id to OpenID Connect Relying Party (client) configuration.
    /// Can't be specified in environment variables.
    pub clients: BTreeMap<String, OidcClientConfiguration>,
}

/// Configuration of how the HTTP interface should listen.
#[derive(Config)]
pub struct ListenConfiguration {
    /// What host and port to bind to. e.g. `127.0.0.1:8177`
    #[config(env = "IC_LISTEN_BIND")]
    pub bind: Option<SocketAddr>,

    /// The public base URI, e.g. `https://idcoop.example.org`.
    /// This is how users will access idCoop.
    /// Please use a HTTPS URI in real-world circumstances.
    pub public_base_uri: String,

    /// How to resolve client IP addresses.
    ///
    /// See [`SecureClientIpSource`] for the options.
    /// The default uses the IP address of the TCP client, which assumes a reverse proxy is not in use.
    /// A popular alternative would be to use the `X-Forwarded-For` header and configure your reverse proxy
    /// (e.g. nginx) to set that.
    #[config(default = "ConnectInfo")]
    pub client_ip_source: SecureClientIpSource,
}

/// Configuration for connecting to Postgres
#[derive(Config)]
pub struct PostgresConfiguration {
    /// Postgres connection string in the form:
    /// `postgres://user:pass@host:port/db_name?application-name=idcoop`
    /// See <https://docs.rs/sqlx-postgres/0.7.2/sqlx_postgres/struct.PgConnectOptions.html>
    #[config(env = "IC_POSTGRES")]
    pub connect: String,
}

/// Configuration for one individual OpenID Connect client (application / relying party).
#[derive(Deserialize)]
pub struct OidcClientConfiguration {
    /// Whether to skip the consent screen. Usually desirable for SSO where you control the service and the identity server.
    /// Defaults to false.
    #[serde(default)]
    pub skip_consent: bool,

    /// OIDC/OAuth2 redirect_uri for the application. Can specify multiple if really needed for some reason.
    pub redirect_uris: BTreeSet<String>,

    /// Friendly name for the service. Will be shown in the user interface.
    pub name: String,

    /// TODO User classes to allow
    /// Must be explicit because it is security sensitive and we don't want a typo to fail open.
    /// User classes are defined by the admin but at the very least includes 'active' and 'not active' (implied if no 'active' class set).
    /// (Design subject to change)
    /// TODO not sure this design is current
    pub allow_user_classes: Vec<String>,

    /// The shared secret for the client.
    /// Must be populated (this is checked on startup).
    /// Can be loaded from a separated secret configuration file.
    /// TODO
    /// - We should consider supporting other auth methods in the future.
    pub secret: Option<String>,
}

/// Content of the secret configuration file; separate from the main configuration file.
#[derive(Config)]
pub struct SeparateSecretConfiguration {
    /// Client secrets for the OIDC clients.
    #[config(default = {})]
    pub oidc_secrets: BTreeMap<String, String>,
}

/// Loaded secrets
pub struct SecretConfig {
    /// The RSA key pair used for OpenID Connect.
    pub rsa_key_pair: RsaKeyPair,
}

impl SecretConfig {
    /// Attempt to load the secrets.
    ///
    /// Fails if we can't read the RSA key for example.
    pub async fn try_new(config: &Configuration) -> eyre::Result<Self> {
        let rsa_keypair_bytes = tokio::fs::read(&config.oidc.rsa_keypair)
            .await
            .context("failed to load RSA private key")?;
        let rsa_key_pair =
            RsaKeyPair::from_pem(rsa_keypair_bytes).context("Failed to decode RSA key pair")?;

        Ok(Self { rsa_key_pair })
    }
}

/// Configuration for the password hashing algorithm. Argon2id is in use.
/// OWASP shows how you can trade-off the different parameters whilst still retaining, to some extent, a similar level of security.
#[derive(Config)]
pub struct PasswordHashingConfig {
    /// Kibibytes of memory to use whilst hashing. Recommendation from OWASP is at least 19 Mebibytes.
    /// I don't want to bump this above 20 MiB in the default setting since we are aiming for low memory use in idCoop.
    /// Very keen users could probably reduce this safely as long as the number of iterations is increased, but consider your own research first.
    #[config(default = 20480)]
    pub memory: u32,
    /// Number of iterations. Recommmendation from OWASP is at least 2. I find 32 plenty fast on my machine; I expect it would be reasonable to bump up more too.
    #[config(default = 32)]
    pub iterations: u32,
    /// Degree of parallelism. Recommendation from OWASP is (at least?) 1.
    #[config(default = 1)]
    pub parallelism: u32,
}

/// Configuration for rate-limits.
#[derive(Config)]
pub struct RatelimitsConfig {
    /// How many login attempts to allow from one source IP address.
    #[config(default = "10 per hour, 5 burst")]
    pub login: RatelimiterConfig,
}

/// Configuration for one rate limit.
///
/// Is read from the configuration file in one of these formats:
/// - "5 Hz, 20 burst"
/// - "5 per second, 10 burst"
/// - "10 per hour, 5 burst"
#[derive(Debug)]
pub struct RatelimiterConfig {
    /// The inner [`Quota`], which this struct is just a wrapper for.
    pub quota: Quota,
}

impl<'de> Deserialize<'de> for RatelimiterConfig {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let stringy_format = String::deserialize(deserializer)?;
        let Some((left, right)) = stringy_format.split_once(',') else {
            return Err(D::Error::custom(
                "no comma. ratelimiter string should be like '5 Hz, 20 burst'",
            ));
        };
        let Some((left_val, left_unit)) = left.trim().split_once(' ') else {
            return Err(D::Error::custom(
                "no units on left. ratelimiter string should be like '5 Hz, 20 burst'",
            ));
        };
        let Some((right_val, right_unit)) = right.trim().split_once(' ') else {
            return Err(D::Error::custom(
                "no units on right. ratelimiter string should be like '5 Hz, 20 burst'",
            ));
        };

        let Ok(left_val) = left_val.parse::<NonZeroU32>() else {
            return Err(D::Error::custom(
                "bad value on left. ratelimiter string should be like '5 Hz, 20 burst'",
            ));
        };
        let Ok(right_val) = right_val.parse::<NonZeroU32>() else {
            return Err(D::Error::custom(
                "bad value on right. ratelimiter string should be like '5 Hz, 20 burst'",
            ));
        };

        let quota = match left_unit.trim() {
            "Hz" | "per second" => Quota::per_second(left_val),
            "per hour" => Quota::per_hour(left_val),
            _other => {
                return Err(D::Error::custom(
                    "bad units on left. ratelimiter string should be like '5 Hz, 20 burst' or '5 per hour, 20 burst'.",
                ));
            }
        };

        if right_unit.trim() != "burst" {
            return Err(D::Error::custom(
                "bad units on right. ratelimiter string should be like '5 Hz, 20 burst' or '5 per hour, 20 burst'.",
            ));
        }

        let quota = quota.allow_burst(right_val);

        Ok(RatelimiterConfig { quota })
    }
}

#[cfg(test)]
mod test {
    use crate::config::RatelimiterConfig;

    #[test]
    fn test_ratelimiter_deser_errors() {
        fn deser(s: &str) -> Result<RatelimiterConfig, serde_json::Error> {
            serde_json::from_value(serde_json::Value::String(s.to_owned()))
        }

        // this is fine
        deser("5 Hz, 20 burst").unwrap();

        // no comma
        assert_eq!(
            deser("5 Hz 20 burst").unwrap_err().to_string(),
            "no comma. ratelimiter string should be like '5 Hz, 20 burst'"
        );

        // bad numbers
        assert_eq!(
            deser("five Hz, 20 burst").unwrap_err().to_string(),
            "bad value on left. ratelimiter string should be like '5 Hz, 20 burst'"
        );
        assert_eq!(
            deser("5 Hz, twenty burst").unwrap_err().to_string(),
            "bad value on right. ratelimiter string should be like '5 Hz, 20 burst'"
        );

        // wrong order
        assert_eq!(
            deser("20 burst, 5 Hz").unwrap_err().to_string(),
            "bad units on left. ratelimiter string should be like '5 Hz, 20 burst' or '5 per hour, 20 burst'."
        );

        // no units
        assert_eq!(
            deser("5 Hz, 20").unwrap_err().to_string(),
            "no units on right. ratelimiter string should be like '5 Hz, 20 burst'"
        );
        assert_eq!(
            deser("5, 20 burst").unwrap_err().to_string(),
            "no units on left. ratelimiter string should be like '5 Hz, 20 burst'"
        );

        // bad units
        assert_eq!(
            deser("20 per milleniumm, 20 burst")
                .unwrap_err()
                .to_string(),
            "bad units on left. ratelimiter string should be like '5 Hz, 20 burst' or '5 per hour, 20 burst'."
        );
        assert_eq!(
            deser("5 Hz, 20 wombats").unwrap_err().to_string(),
            "bad units on right. ratelimiter string should be like '5 Hz, 20 burst' or '5 per hour, 20 burst'."
        );
    }
}
