//! Password handling utilities for idCoop.
//!
//! This file is particularly security-sensitive.

use argon2::{
    password_hash::SaltString, Algorithm, Argon2, Params, PasswordHash, PasswordHasher,
    PasswordVerifier, Version,
};
use eyre::eyre;
use rand::rngs::OsRng;

use crate::config::PasswordHashingConfig;

/// Create a password hash from the given password string and configuration.
///
/// The returned result should only be STORED AS-IS and then used with [`check_hash`] later.
///
/// It is not valid to compare results of this function directly.
pub fn create_password_hash(
    password: &str,
    config: &PasswordHashingConfig,
) -> eyre::Result<String> {
    let salt = SaltString::generate(&mut OsRng);
    let argon2 = Argon2::new(
        Algorithm::Argon2id,
        Version::default(),
        Params::new(config.memory, config.iterations, config.parallelism, None)
            .map_err(|err| eyre!("bad Argon2 params: {err:?}"))?,
    );
    argon2
        .hash_password(password.as_bytes(), &salt)
        .map_err(|err| eyre!("failed argon2 password hash: {err:?}"))
        .map(|h| h.to_string())
}

/// Check a password hash, which was created with [`create_password_hash`] beforehand,
/// against a user-entered password string.
///
/// Under no circumstances should the user be able to control the password hash.
///
/// Returns true if the password is valid.
/// Returns false if the wrong password was supplied.
pub fn check_hash(password: &str, hash: &str) -> eyre::Result<bool> {
    let argon2 = Argon2::new(
        Algorithm::Argon2id,
        Version::default(),
        // the params here are irrelevant: the PasswordHash contains the needed params.
        Params::default(),
    );
    let hash =
        PasswordHash::new(hash).map_err(|err| eyre!("failed to decode password hash: {err:?}"))?;

    Ok(argon2.verify_password(password.as_bytes(), &hash).is_ok())
}

#[cfg(test)]
mod test {
    use rstest::{fixture, rstest};

    use crate::{config::PasswordHashingConfig, passwords::check_hash};

    use super::create_password_hash;

    /// Password hash for "secret"
    const EXAMPLE_SECRET_PASSWORD_HASH: &str = "$argon2id$v=19$m=512,t=1,p=1$Z11PjkMSx/rm4IbDzDmK2Q$VtUH6Iee/GD1FltULyLf6/QRwjNA9d5+mjAKS+WzlZw";

    /// Password hash for "verysecret"
    const EXAMPLE_VERYSECRET_PASSWORD_HASH: &str = "$argon2id$v=19$m=512,t=1,p=1$tcGnJC8EOR4B43z35roeFg$o3grpVK9HAb3850iRI1/nXR+nPZbyFWchmkbkBhm7Co";

    #[fixture]
    fn password_config() -> PasswordHashingConfig {
        PasswordHashingConfig {
            memory: 512,
            iterations: 1,
            parallelism: 1,
        }
    }

    #[rstest]
    fn test_valid_password(password_config: PasswordHashingConfig) {
        let pwh = create_password_hash("secret", &password_config).unwrap();

        assert!(check_hash("secret", &pwh).unwrap());
        assert!(check_hash("secret", EXAMPLE_SECRET_PASSWORD_HASH).unwrap());
    }

    #[rstest]
    fn test_invalid_password(password_config: PasswordHashingConfig) {
        let pwh = create_password_hash("verysecret", &password_config).unwrap();

        assert!(!check_hash("secret", &pwh).unwrap());
        assert!(!check_hash("secret", EXAMPLE_VERYSECRET_PASSWORD_HASH).unwrap());
    }
}
