//! # idCoop database access
//!
//! This file contains PostgreSQL queries.

use chrono::DateTime;
use chrono::NaiveDateTime;
use chrono::Utc;
use eyre::eyre;
use eyre::Context;
use futures::future::BoxFuture;
use sqlx::{types::Uuid, Connection, PgPool, Postgres, Transaction};

use crate::web::login::{
    LoginSession, LOGIN_SESSION_TOKEN_HASH_BYTES, LOGIN_SESSION_XSRF_SECRET_BYTES,
};
use crate::web::oauth_openid::application_session_access_token::ApplicationSession;

/// Postgres-backed storage for IdCoop
/// A connection pool is in use.
pub struct IdCoopStore {
    db_pool: PgPool,
}

impl IdCoopStore {
    /// Connect to a Postgres database using the given Postgres connection URI.
    ///
    /// Once connected, apply database migrations to update the database.
    ///
    /// The simplest example of a connection URI is `postgres:`,
    /// which will connect according to libpq environment variables and defaults.
    ///
    /// A description of the exact behaviour should be sought from the SQLx documentation: [`sqlx::postgres::PgConnectOptions`].
    pub async fn connect(conn_str: &str) -> eyre::Result<Self> {
        // TODO(sqlite) We might support SQLite as well, so we might want to use an 'AnyPool' after
        // installing the requisite drivers here
        let db_pool = PgPool::connect(conn_str)
            .await
            .context("failed to establish Postgres connection")?;

        sqlx::migrate!()
            .run(&db_pool)
            .await
            .context("failed to apply database migrations")?;

        Ok(Self { db_pool })
    }

    /// Run a transaction against the database.
    /// As long as you return an eyre::Result, the return type is up to the caller.
    ///
    /// If the result is Ok(R), the transaction will be committed.
    /// If the result is Err(R), the transaction will NOT be committed.
    pub async fn txn<F, R>(&self, f: F) -> eyre::Result<R>
    where
        for<'a, 'b> F: FnOnce(IdCoopStoreTxn<'a, 'b>) -> BoxFuture<'a, eyre::Result<R>>,
    {
        let mut conn = self
            .db_pool
            .acquire()
            .await
            .context("failed to acquire connection")?;
        let mut txn = conn.begin().await.context("failed to start transaction")?;

        let r = f(IdCoopStoreTxn { txn: &mut txn }).await?;

        txn.commit().await.context("failed to commit transaction")?;

        Ok(r)
    }
}

/// Representation of a user
#[derive(Debug)]
pub struct User {
    /// The unique ID for the user.
    /// Should never change.
    pub user_id: Uuid,
    /// The unique system name for the user.
    /// It is ill-advised to change this but real-world constraints may require it.
    pub user_name: String,
    /// The time when this user was created, in UTC.
    pub created_at_utc: NaiveDateTime,
    /// The password hash of the user. See [`crate::passwords`].
    pub password_hash: Option<String>,
    /// Whether the user is locked and is therefore not allowed to log in.
    pub locked: bool,
}

/// Representation of the action of creating a user.
pub struct CreateUser {
    /// The system name for the user.
    pub user_login_name: String,
    /// The password hash of the user. See [`crate::passwords`].
    pub password_hash: Option<String>,
    /// Whether the user is locked and is therefore not allowed to log in.
    pub locked: bool,
}

/// Basic information about a user
pub struct UserInfo {
    /// The unique system name for the user.
    /// It is ill-advised to change this but real-world constraints may require it.
    pub user_name: String,
    /// The unique ID for the user.
    /// Should never change.
    pub user_id: Uuid,
    /// Whether the user is locked and is therefore not allowed to log in.
    pub locked: bool,
}

/// A wrapper around a database transaction with some database methods on it.
pub struct IdCoopStoreTxn<'a, 'txn> {
    pub(crate) txn: &'a mut Transaction<'txn, Postgres>,
}

impl<'a, 'txn> IdCoopStoreTxn<'a, 'txn> {
    /// Given the hash of an access token and a hash of a refresh token,
    /// invalidates both the access and refresh tokens.
    pub async fn invalidate_access_token_by_hash(
        &mut self,
        access_token_hash: &[u8],
        refresh_token_hash: &[u8],
    ) -> eyre::Result<()> {
        sqlx::query!(
            "
                DELETE FROM application_access_tokens WHERE access_token_hash = $1
            ",
            access_token_hash
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to invalidate access token by hash")?;
        sqlx::query!(
            "
                DELETE FROM application_refresh_tokens WHERE refresh_token_hash = $1
            ",
            refresh_token_hash
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to invalidate refresh token by hash")?;
        Ok(())
    }

    /// Creates a new application session for the given (user, application) pair.
    /// Requires the provided user login session ID to be valid and still active.
    /// Returns a new application session ID, or None if the login session was not active.
    pub async fn create_application_session(
        &mut self,
        user_id: Uuid,
        application_id: &str,
        login_session_id: i32,
    ) -> eyre::Result<Option<i32>> {
        let login_session_opt = sqlx::query!(
            "
                SELECT 1 AS ok FROM login_sessions
                WHERE login_session_id = $1 AND user_id = $2
            ",
            login_session_id,
            user_id
        )
        .fetch_optional(&mut **self.txn)
        .await
        .context("failed to fetch login session when creating application session")?;
        let Some(_login_session) = login_session_opt else {
            return Ok(None);
        };

        let row = sqlx::query!(
            "
                INSERT INTO application_sessions (user_id, application_id, started_at_utc, last_seen_at_utc)
                VALUES ($1, $2, NOW(), NOW())
                RETURNING session_id
            ",
            user_id,
            application_id
        )
        .fetch_one(&mut **self.txn)
        .await
        .context("failed to create application session")?;
        Ok(Some(row.session_id))
    }

    /// Given its hash, issue a new access token.
    ///
    /// The access token is associated with the given session ID and expires at the given time.
    pub async fn issue_access_token(
        &mut self,
        access_token_hash: &[u8],
        session_id: i32,
        expires_at: DateTime<Utc>,
    ) -> eyre::Result<()> {
        let expires_at = expires_at.naive_utc();
        sqlx::query!(
            "
                INSERT INTO application_access_tokens (access_token_hash, session_id, issued_at_utc, expires_at_utc)
                VALUES ($1, $2, NOW(), $3)
            ",
            access_token_hash, session_id, expires_at
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to issue access token")?;
        Ok(())
    }

    /// Given its hash, issue a new refresh token.
    ///
    /// The refresh token is associated with the given session ID and expires at the given time.
    pub async fn issue_refresh_token(
        &mut self,
        refresh_token_hash: &[u8],
        session_id: i32,
        expires_at: DateTime<Utc>,
    ) -> eyre::Result<()> {
        let expires_at = expires_at.naive_utc();
        sqlx::query!(
            "
                INSERT INTO application_refresh_tokens (refresh_token_hash, session_id, issued_at_utc, expires_at_utc)
                VALUES ($1, $2, NOW(), $3)
            ",
            refresh_token_hash, session_id, expires_at
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to issue access token")?;
        Ok(())
    }

    /// Creates a user and returns the user ID.
    pub async fn create_user(&mut self, cu: CreateUser) -> eyre::Result<Uuid> {
        let r = sqlx::query!(
            "INSERT INTO users (user_name, user_id, created_at_utc, password_hash, locked) VALUES ($1, gen_random_uuid(), NOW(), $2, $3) RETURNING user_id",
            &cu.user_login_name, cu.password_hash.as_ref(), cu.locked
        )
        .fetch_one(&mut **self.txn)
        .await.context("failed to create user in DB")?;

        Ok(r.user_id)
    }

    /// Given a user's system name, return their user record if they exist.
    pub async fn lookup_user_by_name(&mut self, name: String) -> eyre::Result<Option<User>> {
        sqlx::query_as!(
            User,
            "SELECT user_id, user_name, password_hash, locked, created_at_utc FROM users WHERE user_name = $1",
            name
        )
        .fetch_optional(&mut **self.txn)
        .await.context("failed to lookup user from DB")
    }

    /// Set a given user's password.
    pub async fn change_user_password(
        &mut self,
        user_id: Uuid,
        new_password_hash: Option<String>,
    ) -> eyre::Result<()> {
        sqlx::query!(
            "UPDATE users SET password_hash = $1 WHERE user_id = $2",
            new_password_hash,
            user_id,
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to set user password in DB")?;
        Ok(())
    }

    /// Set the locked status of a user.
    pub async fn set_user_locked(&mut self, user_id: Uuid, locked: bool) -> eyre::Result<()> {
        sqlx::query!(
            "UPDATE users SET locked = $1 WHERE user_id = $2",
            locked,
            user_id
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to set user (un)locked")?;
        Ok(())
    }

    /// Delete a user.
    ///
    /// Note that this can sometimes be undesirable, as downstream clients won't realise
    /// the system name has been freed.
    pub async fn delete_user(&mut self, user_id: Uuid) -> eyre::Result<()> {
        sqlx::query!("DELETE FROM users WHERE user_id = $1", user_id)
            .execute(&mut **self.txn)
            .await
            .context("failed to delete user")?;
        Ok(())
    }

    /// List all users, in order of their user name, returning brief information about them.
    pub async fn list_user_info(&mut self) -> eyre::Result<Vec<UserInfo>> {
        sqlx::query_as!(
            UserInfo,
            "SELECT user_name, user_id, locked FROM users ORDER BY user_name"
        )
        .fetch_all(&mut **self.txn)
        .await
        .context("failed to list users")
    }

    /// Create a login (~'web UI') session for the given user, with the given hash of the login session's token.
    ///
    /// The XSRF secret is generated specifically for this session
    /// and is used to prevent cross-site request forgery (XSRF) attacks.
    pub async fn create_login_session(
        &mut self,
        login_session_token_hash: &[u8; LOGIN_SESSION_TOKEN_HASH_BYTES],
        user_id: Uuid,
        xsrf_secret: &[u8; LOGIN_SESSION_XSRF_SECRET_BYTES],
    ) -> eyre::Result<()> {
        sqlx::query!(
            "INSERT INTO login_sessions (login_session_token_hash, user_id, started_at_utc, xsrf_secret)
            VALUES ($1, $2, NOW(), $3)",
            login_session_token_hash, user_id, xsrf_secret
        )
        .execute(&mut **self.txn)
        .await
        .context("failed to create login session")?;
        Ok(())
    }

    /// Given the login session's token hash, looks up its corresponding login session.
    pub async fn lookup_login_session(
        &mut self,
        login_session_token_hash: &[u8; LOGIN_SESSION_TOKEN_HASH_BYTES],
    ) -> eyre::Result<Option<LoginSession>> {
        let row_opt = sqlx::query!(
            "
            SELECT user_name, user_id, login_session_id, xsrf_secret
            FROM login_sessions INNER JOIN users USING (user_id)
            WHERE login_session_token_hash = $1
        ",
            login_session_token_hash
        )
        .fetch_optional(&mut **self.txn)
        .await
        .context("failed to lookup login session")?;

        let Some(row) = row_opt else {
            return Ok(None);
        };

        Ok(Some(LoginSession {
            user_name: row.user_name,
            user_id: row.user_id,
            login_session_id: row.login_session_id,
            xsrf_secret: row
                .xsrf_secret
                .try_into()
                .map_err(|_| eyre!("cannot retrieve login session: has invalid XSRF token"))?,
        }))
    }

    /// Given an access token's hash, looks up the corresponding application session.
    pub async fn lookup_application_session(
        &mut self,
        access_token_hash: &[u8; 32],
    ) -> eyre::Result<Option<ApplicationSession>> {
        let row_opt = sqlx::query!(
            "
                SELECT user_name, user_id, session_id
                FROM application_access_tokens
                INNER JOIN application_sessions USING (session_id)
                INNER JOIN users USING (user_id)
                WHERE access_token_hash = $1
            ",
            access_token_hash
        )
        .fetch_optional(&mut **self.txn)
        .await
        .context("failed to lookup application session")?;

        let Some(row) = row_opt else {
            return Ok(None);
        };

        Ok(Some(ApplicationSession {
            application_session_id: row.session_id,
            user_name: row.user_name,
            user_id: row.user_id,
        }))
    }
}
