use std::sync::Arc;

use axum::Router;
use confique::{Config, Partial};
use josekit::jwk::alg::rsa::RsaKeyPair;
use metrics::atomics::AtomicU64;
use pgtemp::PgTempDB;
use serde_json::json;

use crate::{
    config::{Configuration, SecretConfig},
    store::IdCoopStore,
    utils::{Clock, RandGen},
    web::make_router,
};

struct TestSystem {
    database: PgTempDB,
    web: Router,
    config: Arc<Configuration>,
    store: Arc<IdCoopStore>,
    clock: Clock,
}

const RSA_KEY_PAIR_PEM: &[u8] = include_bytes!("tests/keypair.pem");
const RSA_PUBLIC_KEY_PEM: &[u8] = include_bytes!("tests/publickey.crt");

mod test_cli;
mod test_oidc_auth_flow;

async fn basic_system() -> TestSystem {
    let temp_db = pgtemp::PgTempDBBuilder::new()
        .with_dbname("test_idcoop")
        .start_async()
        .await;

    let store = IdCoopStore::connect(&temp_db.connection_uri())
        .await
        .expect("failed to connect to pgtemp db");

    let config_partial: <Configuration as confique::Config>::Partial =
        serde_json::from_value(json!({
            "listen": {
                // Not useful, not actually used in the tests
                "bind": "127.0.0.1:1",
                "public_base_uri": "http://idcoop.example.com",
                "client_ip_source": "RightmostXForwardedFor",
            },
            "postgres": {
                "connect": "postgres://not-used-in-tests"
            },
            "oidc": {
                "issuer": "http://issuer.example.com",
                "rsa_keypair": "not-used-in-tests",
                "clients": {
                    "aclient": {
                        "redirect_uris": [
                            "http://aclient.example.com/redirect"
                        ],
                        "name": "AClient",
                        "allow_user_classes": ["active"],
                        "secret": "secretA",
                    }
                }
            },
            "password_hashing": {
                // Use weak password hash settings; we're not testing Argon2 here,
                // we just want it to be fast.
                "memory": 512,
                "iterations": 1,
            },
            "ratelimits": {
                "login": "3 per hour, 2 burst",
            },
        }))
        .expect("bad test config");
    let config =
        Configuration::from_partial(config_partial.with_fallback(Partial::default_values()))
            .expect("failed to load builtin config");

    let secrets = SecretConfig {
        rsa_key_pair: RsaKeyPair::from_pem(RSA_KEY_PAIR_PEM)
            .expect("failed to decode builtin RSA keypair"),
    };

    let config = Arc::new(config);
    let store = Arc::new(store);
    let clock = Clock(Arc::new(AtomicU64::new(0)));
    let randgen = RandGen::new();
    let router = make_router(
        store.clone(),
        config.clone(),
        Arc::new(secrets),
        clock.clone(),
        randgen,
    )
    .await
    .expect("failed to make router");

    TestSystem {
        database: temp_db,
        web: router,
        config,
        store,
        clock,
    }
}

#[tokio::test]
async fn test_demo() {
    let basic = basic_system().await;
}
