//! Tests the OpenID Connect auth flow

use std::collections::BTreeMap;

use axum::http::StatusCode;
use axum_test_helper::{TestClient, TestResponse};
use insta::assert_yaml_snapshot;

use maplit::btreemap;
use sqlx::types::Uuid;

use crate::{passwords::create_password_hash, tests::basic_system};

async fn dump_resp_text(
    req_name: &str,
    resp: TestResponse,
) -> (StatusCode, BTreeMap<String, String>, String) {
    let status = resp.status();
    // convert headers to a simple B-Tree map so they can be serialised in snapshots
    // easily
    let mut headers: BTreeMap<String, String> = resp
        .headers()
        .clone()
        .into_iter()
        // skip headers that are duplicate keys.
        // We don't care about theme for our purposes.
        .filter_map(|(k, v)| Some((k?.to_string(), v.to_str().unwrap().to_owned())))
        .collect();
    // Remove date because it's not stable across tests!
    headers.remove("date");
    // Remove vary because it has multiple values and we don't want to
    // introduce instability into our tests by only allowing one through.
    headers.remove("vary");
    let text = resp.text().await;
    eprintln!("=== Response for {req_name} ===");
    eprintln!("Status: {status:?}");
    eprintln!("Headers: {headers:#?}");
    eprintln!("Body: {text:?}");
    eprintln!("=== End of response ===");
    (status, headers, text)
}

/// Tests the full flow...
#[tokio::test]
async fn test_full_flow() {
    let sys = basic_system().await;

    let uuid = Uuid::nil();
    let pwhash = create_password_hash("secret", &sys.config.password_hashing).unwrap();
    let _: () = sys
        .store
        .txn(|txn| {
            Box::pin(async move {
                sqlx::query(
                    "INSERT INTO users (user_name, user_id, created_at_utc, password_hash, locked) VALUES ($1, $2, NOW(), $3, $4) RETURNING user_id",
                ).bind("robert").bind(uuid).bind(pwhash).bind(false)
                    .fetch_one(&mut **txn.txn)
                    .await.unwrap();
                Ok(())
            })
        })
        .await
        .unwrap();

    let client = TestClient::new(sys.web);

    ///// These requests are on behalf of the user's browser /////

    const CODE_VERIFIER: &str = "verifying";
    // base64(sha256("verifying"))
    const CODE_CHALLENGE: &str = "LeU9Sprdh-i2mzasKGh8-hmbnmzk48l3Siw390dKY3M";

    // 1. /auth request
    let login_url = format!("/login?then=%2Foidc%2Fauth%3Fscope%3Dopenid%26client_id%3Daclient%26response_type%3Dcode%26state%3Dwombat%26redirect_uri%3Dhttp%3A%252F%252Faclient.example.com%252Fredirect%26code_challenge%3D{CODE_CHALLENGE}%26code_challenge_method%3DS256%26nonce%3Dnoncey");
    let resp = client.get(&format!("/oidc/auth?scope=openid&client_id=aclient&response_type=code&state=wombat&redirect_uri=http:%2F%2Faclient.example.com%2Fredirect&code_challenge={CODE_CHALLENGE}&code_challenge_method=S256&nonce=noncey")).send().await;
    let (status, headers, _text) = dump_resp_text("1. /auth request", resp).await;
    assert_eq!(status, 302);
    assert_eq!(headers.get("location").unwrap(), &login_url);

    // 2. /login request
    let resp = client.get(&login_url).send().await;
    let (status, headers, text) = dump_resp_text("2. /login request", resp).await;
    assert_eq!(status, 200);
    assert_yaml_snapshot!("2/login", (headers, text));

    // 3. /login request with credentials
    let resp = client
        .post(&login_url)
        .form(&btreemap! {
            "username" => "robert",
            "password" => "secret",
            "xsrf" => "HL4qRFKUlBqkrPTvAQ6z-w",
        })
        .header("Cookie", "__Host-SessionlessXsrf=HL4qRFKUlBqkrPTvAQ6z-w")
        // /login is rate-limited by IP source and needs an IP
        .header("X-Forwarded-For", "0.0.0.0")
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("3. /login request with credentials", resp).await;
    assert_eq!(status, 302);
    let auth_loc = headers.get("location").unwrap().to_owned();
    assert_yaml_snapshot!("3/login", (headers, text));

    // 4. /auth request
    let resp = client
        .get(&auth_loc)
        .header(
            "Cookie",
            "__Host-LoginSession=Glh_a6j2xs7ryaJWefPsoW59L7xq6QokAzGh-zEcOxY",
        )
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("4. GET /auth after login", resp).await;
    assert_eq!(status, 200);
    assert_yaml_snapshot!("4/auth", (headers, text));

    sys.clock.set_time(30);

    // 5. /auth request with confirmation
    let resp = client
        .post(&auth_loc)
        .header(
            "Cookie",
            "__Host-LoginSession=Glh_a6j2xs7ryaJWefPsoW59L7xq6QokAzGh-zEcOxY",
        )
        .form(&btreemap! {
            "action" => "accept",
            "xsrf" => "0.JpKyqkWckzF6w6btxX2RXv_MlxgOfoYOZJknydValkc",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("5. POST /auth after confirmation", resp).await;
    assert_eq!(status, 302);
    // Note this snapshot includes a Location: header back to the client application
    assert_yaml_snapshot!("5/auth", (headers, text));

    ///// At this point, we make requests on behalf of the client /////

    // 6. /token request
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code" => "UnLS_bGq0ZB4szozTRCJIG-37ibG08zK",
            "code_verifier" => CODE_VERIFIER,
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("6. POST /token", resp).await;
    assert_eq!(status, 200);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    assert_yaml_snapshot!("6/token", (headers, json));

    // 7. /userinfo request
    let resp = client
        .get("/oidc/userinfo")
        .header(
            "Authorization",
            "Bearer pvgYf08qA_ctEIhMP4DFQzbxjiCx8qfgi4cATwGsH9Q",
        )
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("7. /userinfo", resp).await;
    assert_eq!(status, 200);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    assert_yaml_snapshot!("7/userinfo", (headers, json));
}

#[tokio::test]
async fn test_jwks_endpoint() {
    let sys = basic_system().await;
    let client = TestClient::new(sys.web);
    let resp = client.get("/oidc/jwks").send().await;
    let (status, headers, text) = dump_resp_text("/jwks", resp).await;
    assert_eq!(status, 200);
    assert_yaml_snapshot!((headers, text));
}

#[tokio::test]
async fn test_discovery_endpoint() {
    let sys = basic_system().await;
    let client = TestClient::new(sys.web);
    let resp = client.get("/.well-known/openid-configuration").send().await;
    let (status, headers, text) = dump_resp_text("discovery", resp).await;
    assert_eq!(status, 200);
    assert_yaml_snapshot!((headers, text));
}

#[tokio::test]
async fn test_userinfo_bad_auth() {
    let sys = basic_system().await;
    let client = TestClient::new(sys.web);

    // 1. no auth token
    let resp = client.get("/oidc/userinfo").send().await;
    let (status, headers, text) = dump_resp_text("1. no auth token", resp).await;
    assert_eq!(status, 401);
    assert_yaml_snapshot!("1. no auth token", (headers, text));

    // 2. malformed access token
    let resp = client
        .get("/oidc/userinfo")
        .header("Authorization", "Bearer ++++")
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("2. malformed auth token", resp).await;
    assert_eq!(status, 401);
    assert_yaml_snapshot!("2. malformed auth token", (headers, text));

    // 3. wrong access token
    let resp = client
        .get("/oidc/userinfo")
        .header("Authorization", "Bearer aaaa")
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("3. wrong auth token", resp).await;
    assert_eq!(status, 401);
    assert_yaml_snapshot!("3. wrong auth token", (headers, text));
}

/// Tests error conditions in the /token endpoint
#[tokio::test]
async fn test_token_errors() {
    let sys = basic_system().await;

    let uuid = Uuid::nil();
    let pwhash = create_password_hash("secret", &sys.config.password_hashing).unwrap();
    let _: () = sys
        .store
        .txn(|txn| {
            Box::pin(async move {
                sqlx::query(
                    "INSERT INTO users (user_name, user_id, created_at_utc, password_hash, locked) VALUES ($1, $2, NOW(), $3, $4) RETURNING user_id",
                ).bind("robert").bind(uuid).bind(pwhash).bind(false)
                    .fetch_one(&mut **txn.txn)
                    .await.unwrap();
                Ok(())
            })
        })
        .await
        .unwrap();

    let client = TestClient::new(sys.web);

    ///// These requests are on behalf of the user's browser /////

    const CODE_VERIFIER: &str = "verifying";
    // base64(sha256("verifying"))
    const CODE_CHALLENGE: &str = "LeU9Sprdh-i2mzasKGh8-hmbnmzk48l3Siw390dKY3M";

    // 1. /login request with credentials
    let resp = client
        .post("/login")
        .form(&btreemap! {
            "username" => "robert",
            "password" => "secret",
            "xsrf" => "HL4qRFKUlBqkrPTvAQ6z-w",
        })
        .header("Cookie", "__Host-SessionlessXsrf=HL4qRFKUlBqkrPTvAQ6z-w")
        // /login is rate-limited by IP source and needs an IP
        .header("X-Forwarded-For", "0.0.0.0")
        .send()
        .await;
    let (status, _headers, _text) =
        dump_resp_text("1. /login request with credentials", resp).await;
    assert_eq!(status, 302);

    let auth_loc = format!("/oidc/auth?scope=openid&client_id=aclient&response_type=code&state=wombat&redirect_uri=http:%2F%2Faclient.example.com%2Fredirect&code_challenge={CODE_CHALLENGE}&code_challenge_method=S256&nonce=noncey");

    // 2. /auth request with confirmation
    let resp = client
        .post(&auth_loc)
        .header(
            "Cookie",
            "__Host-LoginSession=HL4qRFKUlBqkrPTvAQ6z-xpYf2uo9sbO68miVnnz7KE",
        )
        .form(&btreemap! {
            "action" => "accept",
            "xsrf" => "0.48qkqIorf3dyk1LgVQwyNT82yDHyqHbXge09Rvfsz8Y",
        })
        .send()
        .await;
    let (status, _headers, _text) = dump_resp_text("2. POST /auth after confirmation", resp).await;
    assert_eq!(status, 302);
    eprintln!("{:?}", _text);

    ///// At this point, we make requests on behalf of the client /////

    // 3. /token request with no code
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code_verifier" => CODE_VERIFIER,
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("3. /token no code", resp).await;
    assert_eq!(status, 400);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    assert_yaml_snapshot!("3/token_no_code", (headers, json));

    // 4. /token request with malformed code (not long enough)
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code" => "aaaa",
            "code_verifier" => CODE_VERIFIER,
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("4. /token malformed code", resp).await;
    assert_eq!(status, 400);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    assert_yaml_snapshot!("4/token_malformed_code", (headers, json));

    // 5. /token request with no code_verifier
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code" => "LRtIBH5rO3O7hwWaF_UkuFJy0v2xqtGQ",
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("5. /token no verifier", resp).await;
    assert_eq!(status, 400);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    assert_yaml_snapshot!("5/token_no_verifier", (headers, json));

    // 6. /token request with wrong code_verifier
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code" => "LRtIBH5rO3O7hwWaF_UkuFJy0v2xqtGQ",
            "code_verifier" => "i'm wrong",
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("6. /token wrong verifier", resp).await;
    assert_eq!(status, 400);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    assert_yaml_snapshot!("6/token_wrong_verifier", (headers, json));
}

/// Tests double-requesting a /token (auth code conflict)
#[tokio::test]
async fn test_token_conflict() {
    let sys = basic_system().await;

    let uuid = Uuid::nil();
    let pwhash = create_password_hash("secret", &sys.config.password_hashing).unwrap();
    let _: () = sys
        .store
        .txn(|txn| {
            Box::pin(async move {
                sqlx::query(
                    "INSERT INTO users (user_name, user_id, created_at_utc, password_hash, locked) VALUES ($1, $2, NOW(), $3, $4) RETURNING user_id",
                ).bind("robert").bind(uuid).bind(pwhash).bind(false)
                    .fetch_one(&mut **txn.txn)
                    .await.unwrap();
                Ok(())
            })
        })
        .await
        .unwrap();

    let client = TestClient::new(sys.web);

    ///// These requests are on behalf of the user's browser /////

    const CODE_VERIFIER: &str = "verifying";
    // base64(sha256("verifying"))
    const CODE_CHALLENGE: &str = "LeU9Sprdh-i2mzasKGh8-hmbnmzk48l3Siw390dKY3M";

    // 1. /login request with credentials
    let resp = client
        .post("/login")
        .form(&btreemap! {
            "username" => "robert",
            "password" => "secret",
            "xsrf" => "HL4qRFKUlBqkrPTvAQ6z-w",
        })
        .header("Cookie", "__Host-SessionlessXsrf=HL4qRFKUlBqkrPTvAQ6z-w")
        // /login is rate-limited by IP source and needs an IP
        .header("X-Forwarded-For", "0.0.0.0")
        .send()
        .await;
    let (status, _headers, _text) =
        dump_resp_text("1. /login request with credentials", resp).await;
    assert_eq!(status, 302);

    let auth_loc = format!("/oidc/auth?scope=openid&client_id=aclient&response_type=code&state=wombat&redirect_uri=http:%2F%2Faclient.example.com%2Fredirect&code_challenge={CODE_CHALLENGE}&code_challenge_method=S256&nonce=noncey");

    // 2. /auth request with confirmation
    let resp = client
        .post(&auth_loc)
        .header(
            "Cookie",
            "__Host-LoginSession=HL4qRFKUlBqkrPTvAQ6z-xpYf2uo9sbO68miVnnz7KE",
        )
        .form(&btreemap! {
            "action" => "accept",
            "xsrf" => "0.48qkqIorf3dyk1LgVQwyNT82yDHyqHbXge09Rvfsz8Y",
        })
        .send()
        .await;
    let (status, _headers, _text) = dump_resp_text("2. POST /auth after confirmation", resp).await;
    assert_eq!(status, 302);
    eprintln!("{:?}", _text);

    ///// At this point, we make requests on behalf of the client /////

    // 3. /token request (successful)
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code" => "LRtIBH5rO3O7hwWaF_UkuFJy0v2xqtGQ",
            "code_verifier" => CODE_VERIFIER,
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, _headers, text) = dump_resp_text("3. POST /token", resp).await;
    assert_eq!(status, 200);
    let json: BTreeMap<String, serde_json::Value> = serde_json::from_str(&text).unwrap();
    let access_token = json
        .get("access_token")
        .unwrap()
        .as_str()
        .unwrap()
        .to_owned();

    // 4. /token request (conflicting)
    let resp = client
        .post("/oidc/token")
        .header("Authorization", "Basic YWNsaWVudDpzZWNyZXRB")
        .form(&btreemap! {
            "code" => "LRtIBH5rO3O7hwWaF_UkuFJy0v2xqtGQ",
            "code_verifier" => CODE_VERIFIER,
            "grant_type" => "authorization_code",
            "redirect_uri" => "http://aclient.example.com/redirect",
        })
        .send()
        .await;
    let (status, headers, text) = dump_resp_text("4. POST /token (conflict)", resp).await;
    assert_eq!(status, 400);
    assert_yaml_snapshot!("4/token_conflict", (headers, text));

    // 5. /userinfo (using the access token that should now have expired)
    let resp = client
        .get("/oidc/userinfo")
        .header("Authorization", format!("Bearer {access_token}"))
        .send()
        .await;
    let (status, _headers, _text) = dump_resp_text("7. /userinfo", resp).await;
    assert_eq!(status, 401);
}
