//! Miscellaneous utilities

#[cfg(not(test))]
pub use self::real_utils::{Clock, RandGen};

#[cfg(test)]
pub use self::test_utils::{Clock, RandGen};

#[cfg(not(test))]
mod real_utils {
    use std::time::{Duration, SystemTime, UNIX_EPOCH};

    use chrono::{DateTime, Utc};
    use rand::{thread_rng, RngCore};

    /// A source of random numbers that can be faked for tests.
    #[derive(Clone)]
    pub struct RandGen;

    impl RngCore for RandGen {
        fn next_u32(&mut self) -> u32 {
            thread_rng().next_u32()
        }

        fn next_u64(&mut self) -> u64 {
            thread_rng().next_u64()
        }

        fn fill_bytes(&mut self, dest: &mut [u8]) {
            thread_rng().fill_bytes(dest)
        }

        fn try_fill_bytes(&mut self, dest: &mut [u8]) -> Result<(), rand::Error> {
            thread_rng().try_fill_bytes(dest)
        }
    }

    /// A source of time that can be faked for tests.
    #[derive(Clone)]
    pub struct Clock;

    impl Clock {
        /// Returns the current time as a `DateTime<Utc>`.
        pub fn now_utc(&self) -> DateTime<Utc> {
            Utc::now()
        }

        /// Returns the current time as a i64 of seconds since the Unix epoch.
        pub fn now_timestamp(&self) -> i64 {
            Utc::now().timestamp()
        }

        /// Sleep until the given timestamp.
        pub async fn sleep_until(&self, until_ts: i64) {
            if until_ts < 0 {
                return;
            }
            let sleep_until = UNIX_EPOCH + Duration::from_secs(until_ts as u64);
            let now = SystemTime::now();
            let sleep_for = sleep_until
                .duration_since(now)
                .unwrap_or(Duration::from_secs(0));
            tokio::time::sleep(sleep_for).await
        }
    }
}

#[cfg(test)]
mod test_utils {
    use std::sync::atomic::AtomicU64;
    use std::sync::Arc;
    use std::sync::Mutex;
    use std::time::Duration;

    use chrono::{DateTime, TimeZone, Utc};
    use rand::{RngCore, SeedableRng};
    use rand_xoshiro::Xoshiro256StarStar;

    #[derive(Clone)]
    pub struct RandGen(Arc<Mutex<Xoshiro256StarStar>>);

    impl RandGen {
        #[allow(clippy::new_without_default)]
        pub fn new() -> RandGen {
            RandGen(Arc::new(Mutex::new(Xoshiro256StarStar::seed_from_u64(
                424242,
            ))))
        }
    }

    impl RngCore for RandGen {
        fn next_u32(&mut self) -> u32 {
            let mut rng = self.0.lock().unwrap();
            rng.next_u32()
        }

        fn next_u64(&mut self) -> u64 {
            let mut rng = self.0.lock().unwrap();
            rng.next_u64()
        }

        fn fill_bytes(&mut self, dest: &mut [u8]) {
            let mut rng = self.0.lock().unwrap();
            rng.fill_bytes(dest)
        }

        fn try_fill_bytes(&mut self, dest: &mut [u8]) -> Result<(), rand::Error> {
            let mut rng = self.0.lock().unwrap();
            rng.try_fill_bytes(dest)
        }
    }

    #[derive(Clone)]
    pub struct Clock(pub Arc<AtomicU64>);

    impl Clock {
        pub fn new_test() -> Self {
            Clock(Arc::new(AtomicU64::new(0)))
        }
        pub fn set_time(&self, new: u64) {
            self.0.store(new, std::sync::atomic::Ordering::Relaxed);
        }

        pub fn now_utc(&self) -> DateTime<Utc> {
            Utc.timestamp_opt(self.0.load(std::sync::atomic::Ordering::Relaxed) as i64, 0)
                .earliest()
                .unwrap()
        }

        pub fn now_timestamp(&self) -> i64 {
            self.0.load(std::sync::atomic::Ordering::Relaxed) as i64
        }

        pub async fn sleep_until(&self, until_ts: i64) {
            if until_ts < 0 {
                return;
            }
            // Wait for time to advance past the requested timestamp
            // TODO write a better test sleep implementation
            while self.now_timestamp() < until_ts {
                tokio::time::sleep(Duration::from_millis(1)).await;
            }
        }
    }
}
