//! # idCoop Web Interface
//!
//! This large module exposes the web/HTTP interface of idCoop.
//!
//! The web interface is used for:
//!
//! - users for logging in and managing their own account
//! - administrators for managing idCoop and their users
//! - applications making OpenID Connect requests to check users

use std::{
    net::{IpAddr, SocketAddr},
    sync::{
        atomic::{AtomicU32, Ordering},
        Arc,
    },
    time::{Duration, SystemTime, UNIX_EPOCH},
};

use axum::{
    extract::ConnectInfo,
    http::{HeaderName, HeaderValue, StatusCode, Uri},
    response::{Html, IntoResponse, Response},
    routing::{get, post},
    Extension, Router,
};
use governor::{clock::QuantaClock, state::keyed::DashMapStateStore, RateLimiter};
use hornbeam::{initialise_template_manager, make_template_manager};
use tower_cookies::CookieManagerLayer;
use tower_http::{cors::CorsLayer, set_header::SetResponseHeaderLayer, trace::TraceLayer};
use tracing::error;

use crate::{
    config::{Configuration, RatelimiterConfig, RatelimitsConfig, SecretConfig},
    store::IdCoopStore,
    utils::{Clock, RandGen},
    web::{
        login::{get_login, post_login, PasswordHashInflightLimiter},
        oauth_openid::{
            authorisation::{oidc_authorisation, post_oidc_authorisation_consent},
            ext_codes::VolatileCodeStore,
            oidc_discovery_configuration, oidc_discovery_webfinger, oidc_jwks, oidc_userinfo,
            token::oidc_token,
        },
    },
};

pub mod login;
pub mod oauth_openid;
pub mod sessionless_xsrf;

make_template_manager! {
    static ref TEMPLATING = {
        default_locale: "en",
    };
}

/// Make an axum `Router` but do not bind it to a port.
/// This exposition allows us to perform integration testing easily.
pub(crate) async fn make_router(
    store: Arc<IdCoopStore>,
    config: Arc<Configuration>,
    secrets: Arc<SecretConfig>,
    clock: Clock,
    randgen: RandGen,
) -> eyre::Result<Router> {
    initialise_template_manager!(TEMPLATING);

    let client_ip_source = config.listen.client_ip_source.clone();
    let ratelimiters = Ratelimiters::new_from_config(&config.ratelimits);

    // TODO(security, v0.0.2): Content-Security-Policy
    let router = Router::new()
        .route("/", get(hello))
        .route(
            // CORS NOT to be made available on this endpoint!
            // Block loading this in a frame!
            "/login",
            get(get_login)
                .post(post_login)
                .layer(SetResponseHeaderLayer::overriding(
                    HeaderName::from_static("x-frame-options"),
                    HeaderValue::from_static("DENY"),
                )),
        )
        .route(
            "/oidc/token",
            post(oidc_token).layer(CorsLayer::permissive().max_age(Duration::from_secs(600))),
        )
        .route(
            "/oidc/userinfo",
            get(oidc_userinfo).layer(CorsLayer::permissive().max_age(Duration::from_secs(600))),
        )
        .route(
            // CORS NOT to be made available on this endpoint!
            // Block loading this in a frame!
            "/oidc/auth",
            get(oidc_authorisation)
                .post(post_oidc_authorisation_consent)
                .layer(SetResponseHeaderLayer::overriding(
                    HeaderName::from_static("x-frame-options"),
                    HeaderValue::from_static("DENY"),
                )),
        )
        .route(
            "/oidc/jwks",
            get(oidc_jwks).layer(CorsLayer::permissive().max_age(Duration::from_secs(600))),
        )
        .route(
            "/.well-known/webfinger",
            get(oidc_discovery_webfinger)
                .layer(CorsLayer::permissive().max_age(Duration::from_secs(600))),
        )
        .route(
            "/.well-known/openid-configuration",
            get(oidc_discovery_configuration)
                .layer(CorsLayer::permissive().max_age(Duration::from_secs(600))),
        )
        .layer(TraceLayer::new_for_http())
        .layer(CookieManagerLayer::new())
        .layer(Extension(config))
        .layer(Extension(secrets))
        .layer(Extension(store))
        // TODO(v0.0.2) this limit should be configurable
        .layer(Extension(Arc::new(PasswordHashInflightLimiter::new(1))))
        .layer(client_ip_source.into_extension())
        .layer(Extension(Arc::new(ratelimiters)))
        .layer(Extension(VolatileCodeStore::new(clock.clone())))
        .layer(Extension(clock))
        .layer(Extension(randgen));

    Ok(router)
}

/// Serves, on the bind address specified, the HTTP service
/// including a user interface and any OAuth, OpenID Connect and custom APIs.
#[cfg(not(test))]
pub async fn serve(
    bind: SocketAddr,
    store: Arc<IdCoopStore>,
    config: Arc<Configuration>,
    secrets: Arc<SecretConfig>,
) -> eyre::Result<()> {
    use eyre::Context;
    use tracing::info;

    let router = make_router(store, config, secrets, Clock, RandGen).await?;

    info!("Listening on {bind:?}");
    axum::Server::try_bind(&bind)
        .context("could not bind listen address")?
        .serve(router.into_make_service_with_connect_info::<SocketAddr>())
        .await?;

    Ok(())
}

async fn hello(ConnectInfo(_): ConnectInfo<SocketAddr>) -> impl IntoResponse {
    // TODO(v0.0.2): landing page
    "idCoop is running!"
}

fn make_login_redirect(then_uri: Uri) -> Response {
    // Use a 302 redirect to the login page
    // This is quite common
    let login_page_uri = match then_uri.path_and_query() {
        Some(path_and_query) => {
            let encoded_params = serde_urlencoded::to_string([("then", path_and_query.as_str())])
                .expect("no reason that URL encode should fail");
            format!("/login?{}", encoded_params)
        }
        None => "/login".to_owned(),
    };

    (
        StatusCode::FOUND,
        [("Location", &login_page_uri)],
        "You need to login. Sending you to the login page.",
    )
        .into_response()
}

/// Wrapper to make `eyre::Error: IntoResponse`.
pub struct InternalEyreError(eyre::Error);

impl From<eyre::Error> for InternalEyreError {
    fn from(value: eyre::Error) -> Self {
        InternalEyreError(value)
    }
}

impl IntoResponse for InternalEyreError {
    fn into_response(self) -> Response {
        error!("internal error processing request: {:?}", self.0);
        (
            StatusCode::INTERNAL_SERVER_ERROR,
            Html("<h1>500 Internal Server Error</h1><p>A fault in our system means that your request could not be handled. Please try again soon or contact the administrator if the issue persists.</p>")
        )
        .into_response()
    }
}

/// A convenient `Result` type for web requests.
pub type WebResult<T> = Result<T, InternalEyreError>;

/// Perform housekeeping on the ratelimiters every 60 seconds.
const HOUSEKEEPING_EVERY: u32 = 60 >> 2;

type IpRateLimiter = RateLimiter<IpAddr, DashMapStateStore<IpAddr>, QuantaClock>;

/// Constructed rate limiters
pub struct Ratelimiters {
    /// The login ratelimiter for `/login`
    pub login: IpRateLimiter,

    /// Time of last housekeeping operation, since Unix epoch, but shifted right by 2 places
    pub last_housekeeping: AtomicU32,
}

impl Ratelimiters {
    /// Create a new bunch of rate limiters from configuration
    pub fn new_from_config(config: &RatelimitsConfig) -> Self {
        Self {
            login: Self::governor_from_config(&config.login),
            last_housekeeping: AtomicU32::new(0),
        }
    }

    fn governor_from_config(cfg: &RatelimiterConfig) -> IpRateLimiter {
        RateLimiter::dashmap(cfg.quota)
    }

    /// Do some housekeeping if it hasn't been done recently.
    pub fn housekeeping(&self) {
        let Ok(now) = SystemTime::now().duration_since(UNIX_EPOCH) else {
            return;
        };

        let now = (now.as_secs() >> 2) as u32;
        let last_housekept = self.last_housekeeping.load(Ordering::SeqCst);

        if last_housekept > now - HOUSEKEEPING_EVERY {
            return;
        }

        if self
            .last_housekeeping
            .compare_exchange(last_housekept, now, Ordering::SeqCst, Ordering::SeqCst)
            .is_err()
        {
            return;
        }

        self.login.retain_recent();
    }
}
