//! `/login` and related utilities

use std::{
    collections::BTreeMap,
    net::IpAddr,
    sync::{Arc, Mutex},
};

use async_trait::async_trait;
use axum::{
    extract::{FromRequestParts, Query},
    headers::Cookie as CookieHeader,
    http::{request::Parts, uri::PathAndQuery, StatusCode},
    response::{Html, IntoResponse, Response},
    Extension, Form, TypedHeader,
};
use axum_client_ip::SecureClientIp;
use base64::{prelude::BASE64_URL_SAFE_NO_PAD, Engine};
use blake2::{digest::Mac, Blake2s256, Blake2sMac256, Digest};
use chrono::{DateTime, Duration, TimeZone, Utc};
use eyre::eyre;
use eyre::{bail, Context, ContextCompat};
use governor::Jitter;
use rand::Rng;
use serde::Deserialize;
use sqlx::types::Uuid;
use tokio::sync::Semaphore;
use tower_cookies::{Cookie, Cookies};
use tracing::error;

use crate::{
    config::{Configuration, PasswordHashingConfig},
    passwords::{check_hash, create_password_hash},
    store::IdCoopStore,
    utils::RandGen,
};

use super::{sessionless_xsrf, Ratelimiters, WebResult};

/// This is the password hash in-flight limiter (PHIL).
/// It is useful to ensure two things:
/// - we only do so many password hashes at once, to prevent memory exhaustion
/// - we don't let one IP address hog all the password hashes (by only doing 1 per source IP at once)
///
/// TODO we should change this a bit to ensure fair queuing! (although I think the semaphore is...)
pub struct PasswordHashInflightLimiter {
    overall: Semaphore,
    per_source_ip: Mutex<BTreeMap<IpAddr, (u8, Arc<Semaphore>)>>,
}

impl PasswordHashInflightLimiter {
    /// Create a new password hashing in-flight limiter.
    ///
    /// - `limit` is the maximum number of allowed concurrent password hashing operations.
    pub fn new(limit: usize) -> Self {
        Self {
            overall: Semaphore::new(limit),
            per_source_ip: Default::default(),
        }
    }

    /// Call a function `f` that is subject to this limit.
    /// The source IP address `src_ip` is the IP address of the client on whose behalf this computation
    /// is being performed.
    ///
    /// A single source IP address will not be allowed to hog more than one slot.
    ///
    /// Some level of fairness is intended here: a requester can't add more requests to the queue
    /// whilst one of its requests is being handled, so other requesters get a fair chance in the queue.
    pub async fn do_limited<R>(
        &self,
        src_ip: IpAddr,
        f: impl FnOnce() -> R,
    ) -> Result<R, &'static str> {
        let source_ip_semaphore = {
            let mut psi = self.per_source_ip.lock().unwrap();
            let entry = psi
                .entry(src_ip)
                .or_insert_with(|| (0, Arc::new(Semaphore::new(1))));
            if entry.0 == u8::MAX {
                return Err("too many waiters");
            }
            entry.0 += 1;
            entry.1.clone()
        };

        let _source_ip_permit = source_ip_semaphore
            .acquire()
            .await
            .map_err(|_| "failed to get per-source-ip semaphore permit (should not happen)")?;

        let _overall_permit = self
            .overall
            .acquire()
            .await
            .map_err(|_| "failed to get overall semaphore permit (should not happen)")?;

        let result = f();

        drop(_overall_permit);
        drop(_source_ip_permit);

        {
            let mut psi = self.per_source_ip.lock().unwrap();
            let entry = psi
                .get_mut(&src_ip)
                .expect("unsound: entry vanished. This is a severe bug");
            assert!(entry.0 > 0, "unsound: per-source-ip counter already 0");
            entry.0 -= 1;

            if entry.0 == 0 {
                psi.remove(&src_ip);
            }
        }

        Ok(result)
    }
}

/// Size of the login token itself
pub const LOGIN_SESSION_TOKEN_BYTES: usize = 32;

/// Size of a Blake2s hash of the login token (which is what we store in the database)
pub const LOGIN_SESSION_TOKEN_HASH_BYTES: usize = 32;

/// Size of the XSRF secret in bytes; this is a Blake2sMac256 key size.
/// TODO the Blake2sMac256 also has a 'salt' and 'personal' item which might give us more bytes to play with and
///      perhaps we should be using that too.
///      e.g. perhaps the persona should be a hash of the user's UUID?
pub const LOGIN_SESSION_XSRF_SECRET_BYTES: usize = 32;

/// Represents a login session, which is effectively just a 'web UI' session for a user.
pub struct LoginSession {
    /// The system name of the user who is logged in to the idCoop web UI
    pub user_name: String,
    /// The UUID of the user who is logged in to the idCoop web UI
    pub user_id: Uuid,
    /// The ID of this login session
    pub login_session_id: i32,
    /// The XSRF (cross-site request forgery) prevention secret of this login session
    pub xsrf_secret: [u8; LOGIN_SESSION_XSRF_SECRET_BYTES],
}

/// XSRF token expiry time is 1 week.
pub const XSRF_TOKEN_EXPIRY_TIME: Duration = Duration::milliseconds(1000 * 86400 * 7);

impl LoginSession {
    /// Generates a XSRF token which is bound to this session and expires 1 week in the future.
    pub fn generate_xsrf_token(&self, now: DateTime<Utc>) -> eyre::Result<String> {
        let now_timestamp = now.timestamp();
        let now_8bytes = now_timestamp.to_be_bytes();

        // TODO this is only using the key, not the salt and persona
        let mac_tag_bytes = Blake2sMac256::new_with_salt_and_personal(&self.xsrf_secret, &[], &[])?
            .chain_update(now_8bytes)
            .finalize()
            .into_bytes();
        let mac_b64 = BASE64_URL_SAFE_NO_PAD.encode(mac_tag_bytes);

        Ok(format!("{now_timestamp}.{mac_b64}"))
    }

    /// Validates a XSRF token to check it is bound to this session and hasn't expired (is less than a week old).
    pub fn validate_xsrf_token(&self, token: &str, now: DateTime<Utc>) -> eyre::Result<()> {
        let (timestamp_str, mac_b64) = token
            .split_once('.')
            .context("XSRF token in wrong format")?;
        let timestamp: i64 = timestamp_str
            .parse()
            .context("timestamp in XSRF token is wrong")?;
        let mac_tag_bytes: Vec<u8> = BASE64_URL_SAFE_NO_PAD
            .decode(mac_b64)
            .context("failed to b64decode the MAC in XSRF token")?;

        let timestamp_8bytes = timestamp.to_be_bytes();

        // TODO this is only using the key, not the salt and persona
        // TODO perhaps the persona should be the user UUID hashed?
        Blake2sMac256::new_with_salt_and_personal(&self.xsrf_secret, &[], &[])?
            .chain_update(timestamp_8bytes)
            .verify_slice(&mac_tag_bytes)
            .context("bad MAC in XSRF token")?;

        // At this point, the MAC is correct. All that's left is to check that the timestamp isn't too old.

        if now.signed_duration_since(
            Utc.timestamp_opt(timestamp, 0)
                .earliest()
                .context("XSRF timestamp not valid")?,
        ) > XSRF_TOKEN_EXPIRY_TIME
        {
            bail!("XSRF token expired.");
        }

        Ok(())
    }
}

#[async_trait]
impl<S> FromRequestParts<S> for LoginSession
where
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let Ok(cookies) = TypedHeader::<CookieHeader>::from_request_parts(parts, state).await
        else {
            return Err((StatusCode::UNAUTHORIZED, "No login session."));
        };
        let Some(cookie_val) = cookies.get("__Host-LoginSession").map(str::to_owned) else {
            return Err((StatusCode::UNAUTHORIZED, "No login session."));
        };
        let Ok(login_session_token) = BASE64_URL_SAFE_NO_PAD.decode(&cookie_val) else {
            return Err((StatusCode::UNAUTHORIZED, "Invalid login session token."));
        };
        if login_session_token.len() != LOGIN_SESSION_TOKEN_BYTES {
            return Err((StatusCode::UNAUTHORIZED, "Invalid login session token."));
        }
        let login_session_token_hash: [u8; LOGIN_SESSION_TOKEN_HASH_BYTES] =
            Blake2s256::digest(&login_session_token).into();

        let db_store = Extension::<Arc<IdCoopStore>>::from_request_parts(parts, state)
            .await
            .expect("no db store; this is a programming error");

        match db_store
            .txn(|mut txn| {
                Box::pin(async move { txn.lookup_login_session(&login_session_token_hash).await })
            })
            .await
        {
            Ok(Some(session)) => Ok(session),
            Ok(None) => {
                return Err((StatusCode::UNAUTHORIZED, "Invalid login session."));
            }
            Err(err) => {
                error!("failed to check login session: {err:?}");
                return Err((
                    StatusCode::INTERNAL_SERVER_ERROR,
                    "A fault occurred when checking your login status. If the issue persists, please contact an administrator.",
                ));
            }
        }

        // TODO do we want a middleware to renew the cookie?
    }
}

/// Query string parameters for `GET /login`
#[derive(Deserialize)]
pub struct LoginQuery {
    /// Which page to follow after the login has completed?
    /// This is the path and query component of a URI.
    ///
    /// Security-wise, this must be 'local' to idCoop;
    /// i.e. we don't allow redirecting users to other places after they log in.
    then: Option<String>,
}

/// `GET /login?then=...`
///
/// If logged in, redirects to `then` (if safe to do so) immediately.
///
/// If not logged in, shows a login form.
pub async fn get_login(
    current_session: Option<LoginSession>,
    Query(query): Query<LoginQuery>,
    cookies: Cookies,
    Extension(mut randgen): Extension<RandGen>,
) -> Response {
    match current_session {
        Some(_session) => make_post_login_redirect(query.then),
        None => {
            let xsrf_token = sessionless_xsrf::get_token(&cookies, &mut randgen);
            Html(format!("<form method='POST'>UN<input type='text' name='username'> PW<input type='password' name='password'> <input type='hidden' name='xsrf' value='{}'><button type='submit'>click here to login</button> (temporary form)</form>", xsrf_token)).into_response()
        }
    }
}

/// Body parameters for `POST /login`
#[derive(Deserialize)]
pub struct PostLoginForm {
    /// User-supplied username
    username: String,
    /// User-supplied password
    password: String,
    /// Hidden XSRF token, used to check this request isn't being made from another site
    /// (i.e. protects against cross-site request forgery)
    xsrf: String,
}

/// Perform a password hash operation, even though it's not needed.
///
/// This aims to prevent side-channel attacks such as timing attacks, to discern the existence of a user.
fn dummy_password_hash(
    password: String,
    password_hash_config: &PasswordHashingConfig,
) -> WebResult<Response> {
    let _hash = create_password_hash(&password, password_hash_config)
        .context("unable to hash password!")?;

    Ok(render_login_retry_form())
}

/// Render the form to retry a login. Currently not implemented.
fn render_login_retry_form() -> Response {
    (StatusCode::UNAUTHORIZED, "Wrong username or password!").into_response() // TODO(ui): this should re-render the login form for another go
}

/// `POST /login?then=...`
///
/// Performs a login attempt.
///
/// The login attempt is rate-limited, to prevent password brute-force attacks.
///
/// The form is XSRF protected using [`crate::web::sessionless_xsrf`]
/// (given we don't have a working login session to use sessionful XSRF protection).
///
/// We prevent requesters from trying to discern if given usernames exist or not;
/// whether or not a user exists, a password hash operation will be performed which takes
/// a certain amount of time.
///
/// The response also does not point out that a username doesn't exist.
///
/// If the login is successful, a new login session is created, which is stored in
/// browser cookies and the user is redirected to what they were trying to access
/// that needed the login in the first place.
#[allow(clippy::too_many_arguments)]
pub async fn post_login(
    Query(query): Query<LoginQuery>,
    cookies: Cookies,
    Extension(store): Extension<Arc<IdCoopStore>>,
    Extension(config): Extension<Arc<Configuration>>,
    Extension(phil): Extension<Arc<PasswordHashInflightLimiter>>,
    SecureClientIp(src_ip): SecureClientIp,
    Extension(ratelimiters): Extension<Arc<Ratelimiters>>,
    Extension(mut randgen): Extension<RandGen>,
    Form(form): Form<PostLoginForm>,
) -> WebResult<Response> {
    ratelimiters.housekeeping();
    ratelimiters
        .login
        .until_key_ready_with_jitter(&src_ip, Jitter::up_to(std::time::Duration::from_secs(1)))
        .await;
    if !sessionless_xsrf::check_token(&cookies, &form.xsrf) {
        // Invalid XSRF token: try again
        return Ok(get_login(None, Query(query), cookies, Extension(randgen)).await);
    }

    // retrieve user details
    // N.B. it may be that there is no user returned in this step, but we need to
    // do a password hash check anyway to avoid user enumeration problems.
    let user_opt = store
        .txn(|mut txn| Box::pin(async move { txn.lookup_user_by_name(form.username).await }))
        .await
        .context("failed to look up user for login")?;

    // verify credentials
    let user = match user_opt {
        Some(user) => match &user.password_hash {
            Some(password_hash) => match phil
                .do_limited(src_ip, || check_hash(&form.password, password_hash))
                .await
                .map_err(|_| eyre!("can't check password"))?
            {
                Ok(is_correct) => {
                    if is_correct {
                        user
                    } else {
                        return Ok(render_login_retry_form());
                    }
                }
                Err(err) => {
                    error!("failed to check password hash: {err:?}");
                    return phil
                        .do_limited(src_ip, || {
                            dummy_password_hash(form.password, &config.password_hashing)
                        })
                        .await
                        .map_err(|_| eyre!("can't check password"))?;
                }
            },
            None => {
                return phil
                    .do_limited(src_ip, || {
                        dummy_password_hash(form.password, &config.password_hashing)
                    })
                    .await
                    .map_err(|_| eyre!("can't check password"))?;
            }
        },
        None => {
            return phil
                .do_limited(src_ip, || {
                    dummy_password_hash(form.password, &config.password_hashing)
                })
                .await
                .map_err(|_| eyre!("can't check password"))?;
        }
    };

    // Generate a login session token and store the hash in our database
    let login_session_token = randgen.gen::<[u8; LOGIN_SESSION_TOKEN_BYTES]>();
    let login_session_token_b64 = BASE64_URL_SAFE_NO_PAD.encode(login_session_token);
    let login_session_token_hash: [u8; LOGIN_SESSION_TOKEN_HASH_BYTES] =
        Blake2s256::digest(login_session_token).into();
    let xsrf_secret = randgen.gen::<[u8; LOGIN_SESSION_XSRF_SECRET_BYTES]>();

    // store session in the database
    store
        .txn(|mut txn| {
            Box::pin(async move {
                txn.create_login_session(&login_session_token_hash, user.user_id, &xsrf_secret)
                    .await
            })
        })
        .await
        .context("failed to store session in database")?;

    cookies.add(
        Cookie::build("__Host-LoginSession", login_session_token_b64.clone())
            .path("/")
            .http_only(true)
            .secure(true)
            .same_site(tower_cookies::cookie::SameSite::Strict)
            .max_age(time::Duration::days(500))
            .finish(),
    );
    Ok(make_post_login_redirect(query.then))
}

/// Make a redirect for once the user has logged in.
///
/// We verify that the redirect URI is just a path and query component,
/// to ensure we don't redirect cross-domain or any silly nonsense like that.
/// This is because cross-domain redirects could be a security hazard.
fn make_post_login_redirect(then: Option<String>) -> Response {
    match then {
        Some(then) => match PathAndQuery::try_from(then) {
            Ok(path_and_query) => (
                StatusCode::FOUND,
                [("Location", path_and_query.as_str())],
                "Logged in. Redirecting you back to what you were doing.",
            )
                .into_response(),
            Err(_) => (StatusCode::BAD_REQUEST, "Bad redirect URL after login.").into_response(),
        },
        None => (
            StatusCode::FOUND,
            [("Location", "/")],
            "Logged in. Sending you to the home page.",
        )
            .into_response(),
    }
}
