//! OpenID Connect / OAuth 2.1 support

use std::sync::Arc;

use axum::{http::StatusCode, response::IntoResponse, Extension, Json};
use josekit::jwk::Jwk;
use serde::Serialize;
use sqlx::types::Uuid;

use crate::config::{Configuration, SecretConfig};

use self::application_session_access_token::ApplicationSession;

pub mod application_session_access_token;
pub mod authorisation;
pub mod ext_codes;
pub mod token;

#[derive(Serialize)]
struct UserInfoResponse {
    sub: Uuid,
    name: String,
    preferred_username: String,
    // TODO more...
}

/// `GET /oidc/userinfo`
pub async fn oidc_userinfo(application_session: ApplicationSession) -> impl IntoResponse {
    Json(UserInfoResponse {
        sub: application_session.user_id,
        name: application_session.user_name.clone(),
        preferred_username: application_session.user_name,
    })
}

#[derive(Serialize)]
struct JsonWebKeySet {
    keys: Vec<Jwk>,
}

/// `GET /oidc/jwks`
pub async fn oidc_jwks(Extension(secrets): Extension<Arc<SecretConfig>>) -> impl IntoResponse {
    let mut jwk = secrets.rsa_key_pair.to_jwk_public_key();
    jwk.set_key_use("sig");
    jwk.set_key_id("thekey");
    jwk.set_algorithm("RS256");

    let key_set = JsonWebKeySet { keys: vec![jwk] };

    (StatusCode::OK, Json(key_set))
}

/// `GET /.well-known/webfinger`
pub async fn oidc_discovery_webfinger() -> impl IntoResponse {
    (StatusCode::NOT_IMPLEMENTED, "NOT YET IMPLEMENTED")
}

/// <https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderMetadata>
#[derive(Serialize)]
struct OidcDiscoveryConfiguration {
    /// REQUIRED. URL using the https scheme with no query or fragment component that the OP asserts as its Issuer Identifier.
    /// If Issuer discovery is supported (see Section 2), this value MUST be identical to the issuer value returned by WebFinger.
    /// This also MUST be identical to the iss Claim value in ID Tokens issued from this Issuer.
    issuer: String,

    /// REQUIRED. URL of the OP's OAuth 2.0 Authorization Endpoint
    authorization_endpoint: String,

    /// URL of the OP's OAuth 2.0 Token Endpoint
    token_endpoint: String,

    /// RECOMMENDED. URL of the OP's UserInfo Endpoint [OpenID.Core]. This URL MUST use the https scheme and MAY contain port, path, and query parameter components.
    userinfo_endpoint: String,

    /// REQUIRED. URL of the OP's JSON Web Key Set [JWK] document.
    /// This contains the signing key(s) the RP uses to validate signatures from the OP.
    /// The JWK Set MAY also contain the Server's encryption key(s), which are used by RPs to encrypt requests to the Server.
    /// When both signing and encryption keys are made available, a use (Key Use) parameter value is REQUIRED for all keys in the referenced JWK Set to indicate each key's intended usage
    /// Although some algorithms allow the same key to be used for both signatures and encryption, doing so is NOT RECOMMENDED, as it is less secure.
    /// The JWK x5c parameter MAY be used to provide X.509 representations of keys provided. When used, the bare key values MUST still be present and MUST match those in the certificate.
    jwks_uri: String,

    // RECOMMENDED. URL of the OP's Dynamic Client Registration Endpoint
    // We don't support Dynamic Client Registration in idCoop (yet?), so no point having this.
    //registration_endpoint: String,
    /// RECOMMENDED. JSON array containing a list of the OAuth 2.0 [RFC6749] scope values that this server supports
    /// The server MUST support the openid scope value.
    /// Servers MAY choose not to advertise some supported scope values even when this parameter is used, although those defined in [OpenID.Core] SHOULD be listed, if supported.
    scopes_supported: Vec<String>,

    /// REQUIRED. JSON array containing a list of the OAuth 2.0 response_type values that this OP supports.
    /// Dynamic OpenID Providers MUST support the code, id_token, and the token id_token Response Type values.
    response_types_supported: Vec<String>,

    /// OPTIONAL. JSON array containing a list of the OAuth 2.0 response_mode values that this OP supports, as specified in OAuth 2.0 Multiple Response Type Encoding Practices [OAuth.Responses].
    /// If omitted, the default for Dynamic OpenID Providers is ["query", "fragment"].
    response_modes_supported: Vec<String>,

    /// OPTIONAL. JSON array containing a list of the OAuth 2.0 Grant Type values that this OP supports.
    /// Dynamic OpenID Providers MUST support the authorization_code and implicit Grant Type values and MAY support other Grant Types.
    /// If omitted, the default value is ["authorization_code", "implicit"].
    grant_types_supported: Vec<String>,

    /// REQUIRED. JSON array containing a list of the Subject Identifier types that this OP supports. Valid types include pairwise and public.
    subject_types_supported: Vec<String>,

    /// REQUIRED. JSON array containing a list of the JWS signing algorithms (alg values) supported by the OP for the ID Token to encode the Claims in a JWT [JWT].
    /// The algorithm RS256 MUST be included.
    /// The value none MAY be supported, but MUST NOT be used unless the Response Type used returns no ID Token from the Authorization Endpoint (such as when using the Authorization Code Flow).
    id_token_signing_alg_values_supported: Vec<String>,
    // ...
    // various OPTIONAL fields omitted here.
    // ...

    // RECOMMENDED. JSON array containing a list of the Claim Names of the Claims that the OpenID Provider MAY be able to supply values for.
    // Note that for privacy or other reasons, this might not be an exhaustive list.
    //claims_supported: Vec<String>,
}

/// `GET /.well-known/openid-configuration`
pub async fn oidc_discovery_configuration(
    Extension(config): Extension<Arc<Configuration>>,
) -> impl IntoResponse {
    // must not include trailing slash
    let base = &config.listen.public_base_uri;
    Json(OidcDiscoveryConfiguration {
        issuer: base.to_string(),
        authorization_endpoint: format!("{base}/oidc/auth"),
        token_endpoint: format!("{base}/oidc/token"),
        userinfo_endpoint: format!("{base}/oidc/userinfo"),
        jwks_uri: format!("{base}/oidc/jwks"),
        scopes_supported: vec!["openid".to_owned()],
        response_types_supported: vec!["code".to_owned()],
        response_modes_supported: vec!["query".to_owned()],
        // TODO should we support 'implicit'?
        // TODO(refresh) should we support 'refresh_token'
        grant_types_supported: vec!["authorization_code".to_owned()],
        subject_types_supported: vec!["public".to_owned()],
        // TODO we should support other types?
        id_token_signing_alg_values_supported: vec!["RS256".to_owned()],
    })
}
