//! Middleware for extracting application sessions (access tokens) from requests.

use std::sync::Arc;

use async_trait::async_trait;
use axum::{
    extract::FromRequestParts,
    headers::{authorization::Bearer, Authorization},
    http::{request::Parts, StatusCode},
    Extension, TypedHeader,
};
use base64::{prelude::BASE64_URL_SAFE_NO_PAD, Engine};
use blake2::{Blake2s256, Digest};
use sqlx::types::Uuid;
use tracing::error;

use crate::store::IdCoopStore;

/// Session between a user and an OpenID Connect client (application / relying party).
pub struct ApplicationSession {
    /// The ID of this session
    pub application_session_id: i32,
    /// The system user name of the user
    pub user_name: String,
    /// The user ID of the user
    pub user_id: Uuid,
}

#[async_trait]
impl<S> FromRequestParts<S> for ApplicationSession
where
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let Ok(TypedHeader(Authorization(bearer))) = TypedHeader::<Authorization<Bearer>>::from_request_parts(parts, state).await else {
            return Err((StatusCode::UNAUTHORIZED, "No access token."));
        };
        let Ok(access_token) = BASE64_URL_SAFE_NO_PAD.decode(bearer.token()) else {
            return Err((
                StatusCode::UNAUTHORIZED,
                "Invalid access token."
            ));
        };
        let access_token_hash: [u8; 32] = Blake2s256::digest(&access_token).into();

        let db_store = Extension::<Arc<IdCoopStore>>::from_request_parts(parts, state)
            .await
            .expect("no db store; this is a programming error");

        match db_store
            .txn(|mut txn| {
                Box::pin(async move { txn.lookup_application_session(&access_token_hash).await })
            })
            .await
        {
            Ok(Some(session)) => Ok(session),
            Ok(None) => {
                return Err((StatusCode::UNAUTHORIZED, "Invalid application session."));
            }
            Err(err) => {
                error!("failed to check application session: {err:?}");
                return Err((
                    StatusCode::INTERNAL_SERVER_ERROR,
                    "A fault occurred when checking your application session. If the issue persists, please contact an administrator.",
                ));
            }
        }
    }
}
