//! `/oidc/auth`

use std::sync::Arc;

use axum::{
    extract::{rejection::QueryRejection, OriginalUri, Query},
    http::{StatusCode, Uri},
    response::{Html, IntoResponse, Response},
    Extension, Form,
};

use eyre::{Context, ContextCompat};

use serde::{Deserialize, Serialize};
use tracing::{error, warn};

use crate::{
    config::{Configuration, OidcClientConfiguration},
    utils::{Clock, RandGen},
    web::{
        login::LoginSession,
        make_login_redirect,
        oauth_openid::ext_codes::{AuthCode, AuthCodeBinding},
    },
};

use super::ext_codes::VolatileCodeStore;

/// Query string parameters for the OIDC authorisation request.
#[derive(Deserialize, Debug)]
pub struct AuthorisationQuery {
    /// REQUIRED. OpenID Connect requests MUST contain the openid scope value.
    /// If the openid scope value is not present, the behavior is entirely unspecified.
    /// Other scope values MAY be present.
    /// Scope values used that are not understood by an implementation SHOULD be ignored.
    /// See Sections 5.4 and 11 for additional scope values defined by this specification.
    scope: String,
    /// REQUIRED. OAuth 2.0 Client Identifier valid at the Authorization Server.
    client_id: String,
    /// REQUIRED. OAuth 2.0 Response Type value that determines the authorization processing flow to be used, including what parameters are returned from the endpoints used.
    /// When using the Authorization Code Flow, this value is code.
    response_type: String,
    /// RECOMMENDED. Opaque value used to maintain state between the request and the callback.
    /// Typically, Cross-Site Request Forgery (CSRF, XSRF) mitigation is done by cryptographically binding the value of this parameter with a browser cookie.
    state: Option<String>,
    /// REQUIRED. Redirection URI to which the response will be sent.
    /// This URI MUST exactly match one of the Redirection URI values for the Client pre-registered at the OpenID Provider,
    /// with the matching performed as described in Section 6.2.1 of [RFC3986] (Simple String Comparison).
    /// When using this flow, the Redirection URI SHOULD use the https scheme; however, it MAY use the http scheme, provided that the Client Type is confidential,
    /// as defined in Section 2.1 of OAuth 2.0, and provided the OP allows the use of http Redirection URIs in this case.
    /// The Redirection URI MAY use an alternate scheme, such as one that is intended to identify a callback into a native application.
    redirect_uri: String,
    /// 'RECOMMENDED OR REQUIRED'. I'm not sure what that means although we treat it as REQUIRED here.
    code_challenge: Option<String>,
    /// OPTIONAL (default = 'plain'). We must support 'plain' and 'S256' schemes here.
    code_challenge_method: Option<String>,
    // OIDC: `nonce` which must be passed all the way through to the ID token
    nonce: Option<String>,
}

/// Sadly this is not specified in the spec, but 384 seems like a real-world practical value.
/// (Some clients have been known to generate nonces at least 300 bytes long!)
/// <https://github.com/openid-certification/oidctest/issues/134>
pub const MAX_NONCE_LENGTH: usize = 384;

/// `GET /oidc/auth`
pub async fn oidc_authorisation(
    query: Result<Query<AuthorisationQuery>, QueryRejection>,
    login_session: Option<LoginSession>,
    Extension(config): Extension<Arc<Configuration>>,
    Extension(code_store): Extension<VolatileCodeStore>,
    Extension(clock): Extension<Clock>,
    Extension(mut randgen): Extension<RandGen>,
    OriginalUri(uri): OriginalUri,
) -> Response {
    let Query(query) = match query {
        Ok(query) => query,
        Err(err) => {
            // TODO(ui) this should be a pretty page
            return (
                StatusCode::BAD_REQUEST,
                format!("TODO bad authorisation request: {err:?}"),
            )
                .into_response();
        }
    };

    if let Some(nonce) = &query.nonce {
        if nonce.len() > MAX_NONCE_LENGTH {
            return (
                StatusCode::BAD_REQUEST,
                format!("Bad authorisation request: Nonce too long (> {MAX_NONCE_LENGTH})"),
            )
                .into_response();
        }
    }

    if !query.scope.split(' ').any(|word| word == "openid") {
        warn!("No OpenID scope requested! Continuing anyway, but beware.");
    }

    let (client_id, client_config) = match validate_authorisation_basics(&query, &config) {
        Ok(x) => x,
        Err(resp) => return resp,
    };

    // If the user isn't logged in, we need to get them to do that first and then come back here.
    let Some(login_session) = login_session else {
        return make_login_redirect(uri);
    };

    // If the application requires consent, then we should ask for that.
    if !client_config.skip_consent {
        return show_consent_page(login_session, client_config, Extension(clock), &config).await;
    }

    // No consent needed: process the authorisation.

    process_authorisation(
        query,
        login_session,
        client_id,
        client_config,
        &config,
        &mut randgen,
        &clock,
        &code_store,
    )
    .await
}

/// Body parameters for OIDC authorisation consent.
#[derive(Deserialize)]
pub struct PostConsentForm {
    action: String,
    xsrf: String,
}

/// `POST /oidc/auth`
#[allow(clippy::too_many_arguments)]
pub async fn post_oidc_authorisation_consent(
    Query(query): Query<AuthorisationQuery>,
    login_session: Option<LoginSession>,
    Extension(config): Extension<Arc<Configuration>>,
    Extension(code_store): Extension<VolatileCodeStore>,
    Extension(clock): Extension<Clock>,
    Extension(mut randgen): Extension<RandGen>,
    OriginalUri(uri): OriginalUri,
    Form(form): Form<PostConsentForm>,
) -> Response {
    let (client_id, client_config) = match validate_authorisation_basics(&query, &config) {
        Ok(x) => x,
        Err(resp) => return resp,
    };

    // If the user isn't logged in, we need to get them to do that first and then come back here.
    let Some(login_session) = login_session else {
        return make_login_redirect(uri);
    };

    if login_session
        .validate_xsrf_token(&form.xsrf, clock.now_utc())
        .is_err()
    {
        // XSRF token is not valid, so show the consent form again...
        return show_consent_page(login_session, client_config, Extension(clock), &config).await;
    }

    match form.action.as_str() {
        "accept" => {
            process_authorisation(
                query,
                login_session,
                client_id,
                client_config,
                &config,
                &mut randgen,
                &clock,
                &code_store,
            )
            .await
        }
        "deny" => fail_authorisation_with_redirect(
            &query,
            client_config,
            AuthorisationRedirectableError::AccessDenied,
            "consent denied".to_owned(),
        ),
        _ => (StatusCode::BAD_REQUEST, "Bad form action").into_response(),
    }
}

/// Returns an error response if the authorisation request is obviously wrong in some way.
/// Otherwise, returns (client_id, client_config).
fn validate_authorisation_basics<'a>(
    query: &AuthorisationQuery,
    config: &'a Configuration,
) -> Result<(String, &'a OidcClientConfiguration), Response> {
    let Some(client_config) = config.oidc.clients.get(&query.client_id) else {
        // TODO(ui) format as pretty page
        return Err((StatusCode::BAD_REQUEST, "TODO bad client_id").into_response());
    };

    if !client_config.redirect_uris.contains(&query.redirect_uri) {
        // TODO(ui) format as pretty page
        return Err((StatusCode::BAD_REQUEST, "TODO bad redirect_uri").into_response());
    }

    if &query.response_type != "code" {
        return Err(fail_authorisation_with_redirect(
            query,
            client_config,
            AuthorisationRedirectableError::UnsupportedResponseType,
            "We only support `code` authorisation responses here.".to_owned(),
        ));
    }
    if query.code_challenge.is_none() {
        return Err(fail_authorisation_with_redirect(
            query,
            client_config,
            AuthorisationRedirectableError::InvalidRequest,
            "`code_challenge` not specified.".to_owned(),
        ));
    }
    match query.code_challenge_method.as_deref() {
        None | Some("S256") | Some("plain") => {
            // OK: supported (None = 'plain')
        }
        _other => {
            return Err(fail_authorisation_with_redirect(
                query,
                client_config,
                AuthorisationRedirectableError::InvalidRequest,
                "`code_challenge_method` is not supported.".to_owned(),
            ));
        }
    }

    Ok((query.client_id.to_owned(), client_config))
}

async fn show_consent_page(
    login_session: LoginSession,
    client_config: &OidcClientConfiguration,
    Extension(clock): Extension<Clock>,
    _config: &Configuration,
) -> Response {
    let xsrf_token = login_session
        .generate_xsrf_token(clock.now_utc())
        .expect("must be able to create a XSRF token");
    Html(format!(
        "hi <u>{}</u>, consent to <u>{}</u>? <form method='POST'><input type='hidden' name='xsrf' value='{}'><button type='submit' name='action' value='accept'>Accept</button> <button type='submit' name='action' value='deny'>Deny</button></form>",
        login_session.user_name, client_config.name, xsrf_token
    ))
    .into_response()
}

/// Proceeds with the final stage of the authorisation procedure.
///
/// Should redirect back to the relying party with either a success or fail response.
/// On success, will create authorisation codes in an in-memory store, ready for redemption.
///
/// Preconditions:
/// - any required consent from the user has now been obtained
/// - query.request_uri has been validated as a safe redirect URI
#[allow(clippy::too_many_arguments)]
async fn process_authorisation(
    query: AuthorisationQuery,
    login_session: LoginSession,
    client_id: String,
    _client_config: &OidcClientConfiguration,
    config: &Configuration,
    randgen: &mut RandGen,
    clock: &Clock,
    code_store: &VolatileCodeStore,
) -> Response {
    assert_eq!(
        &query.response_type, "code",
        "`code` responses only. Others should have been filtered out by this point"
    );
    let code_challenge = query
        .code_challenge
        .expect("already checked to have a code_challenge");

    // Generate a 192-bit random code, which fits into exactly 32 base64 characters.
    // This is an arbitrary choice left to us but I feel a 192-bit value is sufficiently random.
    let code = AuthCode::generate_new_random(randgen);
    let code_base64url = code.to_string();

    // Write down the code and other details in-memory with 10 minute expiry...
    // Limit so we only keep a certain amount per peer IP address.
    code_store.add_redeemable(
        code,
        AuthCodeBinding {
            client_id,
            redirect_uri: query.redirect_uri.clone(),
            nonce: query.nonce,
            code_challenge_method: query
                .code_challenge_method
                .unwrap_or_else(|| String::from("plain")),
            code_challenge,
            user_id: login_session.user_id,
            user_login_session_id: login_session.login_session_id,
        },
        clock.now_timestamp() + 600,
    );

    #[derive(Serialize)]
    struct CodeResponseRedirect {
        /// Opaque authorisation code generated by us.
        /// Must expire after a short time (10 min recommended)
        /// Client MUST NOT use it more than once. (If they do, we should revoke all issued refresh tokens and access tokens from that code.)
        /// The auth code is bound to the (client_id, code_challenge, request_uri)
        code: String,

        /// REQUIRED if they sent one to us, otherwise omit.
        /// A verbatim copy of what they sent to us.
        #[serde(skip_serializing_if = "Option::is_none")]
        state: Option<String>,

        /// OPTIONAL but I will treat this as recommended: seems to me there is no harm
        /// Our identifier as an issuer, which the client can use to prevent mixup attacks.
        iss: String,
    }
    let func = move || -> eyre::Result<Response> {
        let redirect_uri = extend_uri_with_params(
            query
                .redirect_uri
                .parse()
                .context("redirect_uri unparseable")?,
            CodeResponseRedirect {
                code: code_base64url,
                state: query.state,
                iss: config.oidc.issuer.clone(),
            },
        )
        .context("failed to build extended URI")?
        .to_string();

        Ok((
            StatusCode::FOUND,
            [("Location", redirect_uri.as_str())],
            "Authorisation succeeded; redirecting you back.",
        )
            .into_response())
    };

    func().unwrap_or_else(|e| {
        error!("failed to redirect failed auth: {e:?}");
        (StatusCode::INTERNAL_SERVER_ERROR, "Internal server error").into_response()
    })
}

/// Fails the authorisation procedure and redirects back to the Relying Party.
///
/// Preconditions:
/// - query.redirect_uri has been validated as a safe redirect URI
fn fail_authorisation_with_redirect(
    query: &AuthorisationQuery,
    _client_config: &OidcClientConfiguration,
    error: AuthorisationRedirectableError,
    error_description: String,
) -> Response {
    #[derive(Serialize)]
    struct ErrorRedirect {
        error: AuthorisationRedirectableError,
        error_description: String,
    }
    let func = move || -> eyre::Result<Response> {
        let redirect_uri = extend_uri_with_params(
            query
                .redirect_uri
                .parse()
                .context("redirect_uri unparseable")?,
            ErrorRedirect {
                error,
                error_description,
            },
        )
        .context("failed to build extended URI")?
        .to_string();

        Ok((
            StatusCode::FOUND,
            [("Location", redirect_uri.as_str())],
            "Authorisation failed; redirecting you back.",
        )
            .into_response())
    };

    func().unwrap_or_else(|e| {
        error!("failed to redirect failed auth: {e:?}");
        (StatusCode::INTERNAL_SERVER_ERROR, "Internal server error").into_response()
    })
}

fn extend_uri_with_params(uri: Uri, params: impl Serialize) -> eyre::Result<Uri> {
    let mut uri_parts = uri.into_parts();
    let paq = uri_parts
        .path_and_query
        .as_mut()
        .context("TODO empty path and query not supported for client redirection endpoints yet")?;
    let mut query = paq.query().unwrap_or("").to_owned();
    let appendage = serde_urlencoded::to_string(params).context("failed to serialise extension")?;
    if query.is_empty() {
        query.push('?');
    } else {
        query.push('&');
    }
    query.push_str(&appendage);
    uri_parts.path_and_query = Some(format!("{}{}", paq.path(), query).parse()?);
    Uri::from_parts(uri_parts).context("failed to rebuild URI from parts")
}

// pub async fn oidc_authorisation_action(
//     query: Result<Query<AuthorisationQuery>, QueryRejection>,
//     Extension(config): Extension<Arc<Configuration>>,
// )

/// Errors in the authorisation procedure for which it is acceptable to redirect back to the Relying Party.
///
/// OAuth 2.1: https://datatracker.ietf.org/doc/html/draft-ietf-oauth-v2-1-09#name-error-response-2
///
/// Errors beginning with 'Oidc' are from
/// OpenID Connect Core: https://openid.net/specs/openid-connect-core-1_0.html#AuthError
/// (Not all errors from OpenID Connect Core were included.)
#[derive(Copy, Clone, Serialize, Deserialize)]
#[allow(missing_docs)]
pub enum AuthorisationRedirectableError {
    #[serde(rename = "invalid_request")]
    InvalidRequest,
    #[serde(rename = "unauthorized_client")]
    UnauthorisedClient,
    #[serde(rename = "access_denied")]
    AccessDenied,
    #[serde(rename = "unsupported_response_type")]
    UnsupportedResponseType,
    #[serde(rename = "invalid_scope")]
    InvalidScope,
    #[serde(rename = "server_error")]
    ServerError,
    #[serde(rename = "temporarily_unavailable")]
    TemporarilyUnavailable,
    #[serde(rename = "interaction_required")]
    OidcInteractionRequired,
    #[serde(rename = "login_required")]
    OidcLoginRequired,
    #[serde(rename = "account_selection_required")]
    OidcAccountSelectionRequired,
    #[serde(rename = "consent_required")]
    OidcConsentRequired,
}
