//! Authorisation codes

use std::{
    collections::{BTreeSet, HashMap},
    fmt::Debug,
    fmt::Display,
    str::FromStr,
    sync::{Arc, Mutex},
};

use base64::{display::Base64Display, prelude::BASE64_URL_SAFE_NO_PAD, Engine};
use rand::Rng;
use sqlx::types::Uuid;
use tokio::sync::Notify;

use crate::utils::{Clock, RandGen};

/// Display shows the auth code as base64 (URL-safe non-padded).
/// FromStr/parse parses the same format.
#[derive(Clone, Hash, PartialEq, Eq, Ord, PartialOrd)]
pub struct AuthCode(pub [u8; 24]);

/// Access token
pub type AccessToken = [u8; 32];
/// Hash of an access token
pub type AccessTokenHash = [u8; 32];
/// Refresh token
pub type RefreshToken = [u8; 32];
/// Hash of a refresh token
pub type RefreshTokenHash = [u8; 32];

impl Display for AuthCode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            Base64Display::new(&self.0, &BASE64_URL_SAFE_NO_PAD)
        )
    }
}

impl Debug for AuthCode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{self}")
    }
}

impl FromStr for AuthCode {
    type Err = &'static str;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut out = [0u8; 24];
        let read = BASE64_URL_SAFE_NO_PAD
            .decode_slice(s, &mut out)
            .map_err(|_| "wrong size")?;
        if read != 24 {
            return Err("wrong size");
        }
        Ok(AuthCode(out))
    }
}

impl AuthCode {
    /// Generate a new authorisation code using the thread's RNG
    pub fn generate_new_random(randgen: &mut RandGen) -> Self {
        Self(randgen.gen::<[u8; 24]>())
    }
}

/// Binding between an authorisation code (ready to be redeemed)
/// and both the user that authenticated to produce it
/// as well as the OpenID Connect client that it is for.
#[derive(PartialEq, Eq, Debug)]
pub struct AuthCodeBinding {
    /// ID of the OpenID Connect client
    pub client_id: String,
    /// Redirect URI that was used to send this authorisation code back to the OpenID Connect
    /// client. The client must present the same Redirect URI when redeeming the code.
    /// TODO(memory): would be nice to intern this, since it is likely to be repeated a lot...
    pub redirect_uri: String,
    /// A 'number used once', as was specified in the authorisation request
    /// Technically optional in the protocol but it's good practice to use one!
    pub nonce: Option<String>,
    /// The method for using the code challenge
    pub code_challenge_method: String,
    /// The code challenge
    pub code_challenge: String,
    /// The UUID of the user that authenticated to produce this authorisation code
    pub user_id: Uuid,
    /// The login session ID of the user that authenticated.
    /// The authorisation code will not be considered valid when the login session is no longer valid.
    pub user_login_session_id: i32,
}

/// The representation of an auth code that was redeemed by a client.
/// The auth code is not allowed to be doubly redeemed, so by remembering this,
/// we can revoke the tokens.
pub struct RedeemedAuthCode {
    /// The access token hash of whoever redeemed the auth code.
    /// Is used to invalidate the access token if an auth code is double-redeemed.
    access_token_hash: AccessTokenHash,
    /// The refresh token hash of the redeemer.
    refresh_token_hash: RefreshTokenHash,
}

#[derive(Default)]
struct VolatileCodeStoreInner {
    /// Codes that are currently redeemable
    pub redeemable_codes: HashMap<AuthCode, AuthCodeBinding>,

    /// Codes that have been redeemed but still have not expired:
    /// they can still be 'double-claimed' and lead to a conflict being detected.
    pub conflictable_codes: HashMap<AuthCode, RedeemedAuthCode>,

    /// Time when codes will expire
    pub expire_codes_at: BTreeSet<(i64, AuthCode)>,
}

impl VolatileCodeStoreInner {
    pub fn redeem(
        &mut self,
        auth_code: &AuthCode,
        access_token_hash: AccessTokenHash,
        refresh_token_hash: RefreshTokenHash,
    ) -> CodeRedemption {
        if let Some(conflicted) = self.conflictable_codes.get(auth_code) {
            return CodeRedemption::Conflicted {
                access_token_to_invalidate: conflicted.access_token_hash,
                refresh_token_to_invalidate: conflicted.refresh_token_hash,
            };
        }

        let Some((auth_code, binding)) = self.redeemable_codes.remove_entry(auth_code) else {
            return CodeRedemption::Invalid;
        };

        self.conflictable_codes.insert(
            auth_code,
            RedeemedAuthCode {
                access_token_hash,
                refresh_token_hash,
            },
        );

        CodeRedemption::Valid { binding }
    }

    pub fn add_redeemable(
        &mut self,
        auth_code: AuthCode,
        auth_code_binding: AuthCodeBinding,
        expires_at: i64,
    ) {
        self.redeemable_codes
            .insert(auth_code.clone(), auth_code_binding);
        self.expire_codes_at.insert((expires_at, auth_code));
    }

    /// Removes all expired auth codes and returns the time of the earliest next expiry, if present.
    pub(self) fn handle_expiry(&mut self, now: i64) -> Option<i64> {
        loop {
            let (ts, _auth_code) = self.expire_codes_at.first()?;

            // Remove if expired
            if *ts <= now {
                let (_, auth_code) = self.expire_codes_at.pop_first().unwrap();
                self.redeemable_codes.remove(&auth_code);
                self.conflictable_codes.remove(&auth_code);
                continue;
            }

            // Otherwise this is our stopping point
            return Some(*ts);
        }
    }
}

/// Stores authorisation codes, but is volatile (not persisted to disk).
/// This is OK because they should be used practically immediately anyway.
#[derive(Clone)]
pub struct VolatileCodeStore {
    poke: Arc<Notify>,
    inner: Arc<Mutex<VolatileCodeStoreInner>>,
}

impl VolatileCodeStore {
    /// Create a new instance.
    pub fn new(clock: Clock) -> Self {
        let poke = Arc::new(Notify::new());
        let inner: Arc<Mutex<VolatileCodeStoreInner>> = Default::default();

        {
            let poke = poke.clone();
            let inner = inner.clone();
            tokio::spawn(Self::expirer(inner, poke, clock));
        }

        VolatileCodeStore { inner, poke }
    }
}

impl VolatileCodeStore {
    async fn expirer(inner: Arc<Mutex<VolatileCodeStoreInner>>, poke: Arc<Notify>, clock: Clock) {
        let mut next_expiry: Option<i64> = None;
        loop {
            match next_expiry {
                Some(next_expiry) => {
                    let sleep_future = clock.sleep_until(next_expiry);
                    tokio::select! {
                        _ = poke.notified() => {},
                        _ = sleep_future => {},
                    }
                }
                None => {
                    poke.notified().await;
                }
            }

            next_expiry = {
                let mut inner = inner.lock().unwrap();
                inner.handle_expiry(clock.now_timestamp())
            };
        }
    }

    /// Redeem an authorisation code.
    ///
    /// See [`CodeRedemption`] for a description of the possible outcomes here.
    pub fn redeem(
        &self,
        auth_code: &AuthCode,
        access_token_hash: AccessTokenHash,
        refresh_token_hash: RefreshTokenHash,
    ) -> CodeRedemption {
        let mut inner = self.inner.lock().unwrap();
        inner.redeem(auth_code, access_token_hash, refresh_token_hash)
    }

    /// Add a new redeemable authorisation code.
    pub fn add_redeemable(&self, auth_code: AuthCode, binding: AuthCodeBinding, expires_at: i64) {
        let mut inner = self.inner.lock().unwrap();
        inner.add_redeemable(auth_code, binding, expires_at);
        drop(inner);
        // notify expirer that we have inserted something at `expires_at`.
        self.poke.notify_one();
    }
}

/// Possible outcomes of an attempt to redeem an authorisation code.
#[derive(PartialEq, Eq, Debug)]
pub enum CodeRedemption {
    /// That auth code was not active
    Invalid,
    /// That auth code is valid and has been redeemed.
    Valid {
        /// The binding, which describes what this authorisation code is for.
        binding: AuthCodeBinding,
    },
    /// That auth code had already been redeemed: please invalidate the given access token and reject this redemption.
    Conflicted {
        /// The hash of the access token that already redeemed this.
        /// The access token should be invalidated.
        access_token_to_invalidate: AccessTokenHash,
        /// The hash of the refresh token that already redeemed this.
        /// The refresh token should be invalidated.
        ///
        /// TODO(refresh) what if the token was refreshed since?
        refresh_token_to_invalidate: RefreshTokenHash,
    },
}

#[cfg(test)]
mod test {
    use std::{str::FromStr, time::Duration};

    use assert_matches2::assert_matches;
    use rstest::{fixture, rstest};
    use sqlx::types::Uuid;

    use crate::{utils::Clock, web::oauth_openid::ext_codes::CodeRedemption};

    use super::{AuthCode, AuthCodeBinding, VolatileCodeStore, VolatileCodeStoreInner};

    const VALID_CODE: AuthCode = AuthCode([21; 24]);

    const USER_UUID: Uuid = Uuid::nil();
    const USER_LOGIN_SESSION_ID: i32 = 1347;

    #[fixture]
    fn code_store() -> VolatileCodeStoreInner {
        let mut vcs = VolatileCodeStoreInner::default();

        vcs.add_redeemable(
            VALID_CODE.clone(),
            AuthCodeBinding {
                client_id: "client_id".to_owned(),
                redirect_uri: "https://client/redirect".to_owned(),
                nonce: None,
                code_challenge_method: "method".to_owned(),
                code_challenge: "challenge".to_owned(),
                user_id: USER_UUID,
                user_login_session_id: USER_LOGIN_SESSION_ID,
            },
            128,
        );
        vcs
    }

    #[rstest]
    fn test_redeem_nonexistent_code(mut code_store: VolatileCodeStoreInner) {
        let ac = AuthCode([0; 24]);

        let result = code_store.redeem(&ac, [42; 32], [43; 32]);

        assert_eq!(result, CodeRedemption::Invalid);
    }

    #[rstest]
    fn test_redeem_real_code(mut code_store: VolatileCodeStoreInner) {
        let result = code_store.redeem(&VALID_CODE, [42; 32], [43; 32]);

        assert_matches!(result, CodeRedemption::Valid { binding });

        assert_eq!(binding.client_id, "client_id");
    }

    #[rstest]
    fn test_redeem_code_twice(mut code_store: VolatileCodeStoreInner) {
        // redeem a first time
        let result = code_store.redeem(&VALID_CODE, [42; 32], [43; 32]);
        assert_matches!(result, CodeRedemption::Valid { .. });

        // redeem a second time
        let result = code_store.redeem(&VALID_CODE, [1; 32], [2; 32]);
        assert_matches!(
            result,
            CodeRedemption::Conflicted {
                access_token_to_invalidate,
                refresh_token_to_invalidate
            }
        );

        assert_eq!(access_token_to_invalidate, [42; 32]);
        assert_eq!(refresh_token_to_invalidate, [43; 32]);
    }

    #[rstest]
    fn test_expire_and_redeem_not_expired_yet(mut code_store: VolatileCodeStoreInner) {
        code_store.handle_expiry(127);

        // The code shouldn't expire yet.
        let result = code_store.redeem(&VALID_CODE, [42; 32], [43; 32]);
        assert_matches!(result, CodeRedemption::Valid { .. });
    }

    #[rstest]
    fn test_expire_and_redeem_expired_now(mut code_store: VolatileCodeStoreInner) {
        code_store.handle_expiry(128);

        // The code should have expired
        let result = code_store.redeem(&VALID_CODE, [42; 32], [43; 32]);
        assert_eq!(result, CodeRedemption::Invalid);
    }

    #[rstest]
    fn test_expire_and_redeem_conflict(mut code_store: VolatileCodeStoreInner) {
        // redeem a first time
        let result = code_store.redeem(&VALID_CODE, [42; 32], [43; 32]);
        assert_matches!(result, CodeRedemption::Valid { .. });

        code_store.handle_expiry(127);

        // redeem a second time: the conflict should not have expired yet.
        let result = code_store.redeem(&VALID_CODE, [1; 32], [2; 32]);
        assert_matches!(result, CodeRedemption::Conflicted { .. });

        // redeem a third time to show the conflict doesn't get removed.
        let result = code_store.redeem(&VALID_CODE, [1; 32], [2; 32]);
        assert_matches!(result, CodeRedemption::Conflicted { .. });

        code_store.handle_expiry(128);

        // The code and its conflict entry should have expired
        let result = code_store.redeem(&VALID_CODE, [42; 32], [43; 32]);
        assert_eq!(result, CodeRedemption::Invalid);
    }

    #[test]
    fn test_authcode_display() {
        assert_eq!(VALID_CODE.to_string(), "FRUVFRUVFRUVFRUVFRUVFRUVFRUVFRUV");
        assert_eq!(
            format!("{VALID_CODE:?}"),
            "FRUVFRUVFRUVFRUVFRUVFRUVFRUVFRUV"
        );
    }

    #[test]
    fn test_authcode_from_string() {
        assert_eq!(
            AuthCode::from_str(&VALID_CODE.to_string()).unwrap(),
            VALID_CODE
        );
    }

    #[test]
    fn test_authcode_from_string_wrong_size() {
        // Not a valid base64 string
        assert_eq!(AuthCode::from_str("a").unwrap_err(), "wrong size");

        // Not 24 bytes when decoded
        assert_eq!(AuthCode::from_str("abcd").unwrap_err(), "wrong size");
    }

    /// Basic test for the full-fat VolatileCodeStore, not just its inner logic
    /// We can't easily cover everything here but may as well test the basics.
    #[tokio::test]
    async fn test_fullfat_store_basic() {
        let clock = Clock::new_test();
        let vcs = VolatileCodeStore::new(clock.clone());

        vcs.add_redeemable(
            VALID_CODE.clone(),
            AuthCodeBinding {
                client_id: "client_id".to_owned(),
                redirect_uri: "https://client/redirect".to_owned(),
                nonce: None,
                code_challenge_method: "method".to_owned(),
                code_challenge: "challenge".to_owned(),
                user_id: USER_UUID,
                user_login_session_id: USER_LOGIN_SESSION_ID,
            },
            i64::MAX,
        );

        assert_matches!(
            vcs.redeem(&VALID_CODE, [1; 32], [2; 32]),
            CodeRedemption::Valid { .. }
        );

        vcs.add_redeemable(
            VALID_CODE.clone(),
            AuthCodeBinding {
                client_id: "client_id".to_owned(),
                redirect_uri: "https://client/redirect".to_owned(),
                nonce: None,
                code_challenge_method: "method".to_owned(),
                code_challenge: "challenge".to_owned(),
                user_id: USER_UUID,
                user_login_session_id: USER_LOGIN_SESSION_ID,
            },
            // Expiry in the past: this should be auto-expired when poked and
            // given a moment
            1,
        );
        clock.set_time(2);

        vcs.add_redeemable(
            AuthCode([0; 24]),
            AuthCodeBinding {
                client_id: "client_id".to_owned(),
                redirect_uri: "https://client/redirect".to_owned(),
                nonce: None,
                code_challenge_method: "method".to_owned(),
                code_challenge: "challenge".to_owned(),
                user_id: USER_UUID,
                user_login_session_id: USER_LOGIN_SESSION_ID,
            },
            i64::MAX,
        );

        // Give a short time for the expiry to take place.
        tokio::time::sleep(Duration::from_millis(2)).await;

        assert_matches!(
            vcs.redeem(&VALID_CODE, [1; 32], [2; 32]),
            CodeRedemption::Invalid
        );
    }
}
