//! `/oidc/token`

use std::{str::FromStr, sync::Arc};

use axum::{
    extract::rejection::FormRejection,
    headers::{authorization::Basic, Authorization},
    http::StatusCode,
    response::IntoResponse,
    Extension, Form, Json, TypedHeader,
};
use base64::{prelude::BASE64_URL_SAFE_NO_PAD, Engine};
use blake2::Blake2s256;
use chrono::Duration;
use eyre::{bail, Context};
use josekit::{
    jws::{alg::rsassa::RsassaJwsAlgorithm::Rs256, JwsHeader},
    jwt::JwtPayload,
};
use rand::Rng;
use serde::{Deserialize, Serialize};
use sha2::{Digest, Sha256};
use subtle::ConstantTimeEq;
use tracing::{debug, error};

use crate::{
    config::{Configuration, SecretConfig},
    store::IdCoopStore,
    utils::{Clock, RandGen},
};

use super::ext_codes::{
    AccessToken, AccessTokenHash, AuthCode, CodeRedemption, RefreshToken, RefreshTokenHash,
    VolatileCodeStore,
};

/// Body parameters for `POST /token`
#[derive(Deserialize)]
pub struct TokenFormParams {
    grant_type: String,
    redirect_uri: String,
    code: Option<String>,
    code_verifier: Option<String>,
    // /// Used for authentication sometimes if another authentication method is not in use.
    // client_id: Option<String>,
}

/// `POST /oidc/token`
///
/// OpenID Connect clients call this to exchange an authorisation code they received for an access token.
///
/// TODO auth_header can be one alternative auth method
#[allow(clippy::too_many_arguments)]
pub async fn oidc_token(
    basic_auth: Option<TypedHeader<Authorization<Basic>>>,
    Extension(config): Extension<Arc<Configuration>>,
    Extension(secrets): Extension<Arc<SecretConfig>>,
    Extension(store): Extension<Arc<IdCoopStore>>,
    Extension(code_store): Extension<VolatileCodeStore>,
    Extension(mut randgen): Extension<RandGen>,
    Extension(clock): Extension<Clock>,
    form: Result<Form<TokenFormParams>, FormRejection>,
) -> impl IntoResponse {
    let form = match form {
        Ok(form) => form,
        Err(err) => {
            return (
                StatusCode::BAD_REQUEST,
                Json(TokenError {
                    code: TokenErrorCode::InvalidRequest,
                    description: match err {
                        FormRejection::InvalidFormContentType(_) => {
                            "Wrong Content-Type for the /token endpoint".to_owned()
                        }
                        FormRejection::FailedToDeserializeForm(_) => {
                            "Failed to deserialise form".to_owned()
                        }
                        FormRejection::FailedToDeserializeFormBody(err) => {
                            format!("Failed to deserialise form body: {err:?}")
                        }
                        FormRejection::BytesRejection(err) => {
                            format!("Failed to read bytes: {err:?}")
                        }
                        _ => format!("other unhandled problem: {err:?}"),
                    },
                }),
            )
                .into_response();
        }
    };

    // Check client auth (only one auth allowed in any one request to prevent confusion)
    // WE ONLY SUPPORT `client_secret_basic` for the time being
    // TODO we should make this logic reusable for other endpoints
    let Some(TypedHeader(Authorization(basic_auth))) = basic_auth else {
        // TODO(out of spec): we must respond with 401 and WWW-Authenticate if the client tried any Authorization header
        // (See OAuth2.1 3.2.3.1).
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidClient,
                description: "idCoop only supports `client_secret_basic` authorisation at this time and no basic auth was available.".to_string(),
            })
        ).into_response();
    };

    let Some(unverified_client_config) = config.oidc.clients.get(basic_auth.username()) else {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidClient,
                description: "That `client_id` is not recognised here.".to_string(),
            }),
        )
            .into_response();
    };

    if !bool::from(
        basic_auth.password().as_bytes().ct_eq(
            unverified_client_config
                .secret
                .as_ref()
                .expect("client secret missing; should have been checked at startup")
                .as_bytes(),
        ),
    ) {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidClient,
                description: "Failed to authenticate as that client.".to_string(),
            }),
        )
            .into_response();
    }
    // now we have checked that the client is who they say they are...
    let _client_config = unverified_client_config;

    // TODO(refresh) support other grant types, e.g. refresh tokens
    if form.grant_type != "authorization_code" {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::UnsupportedGrantType,
                description: "idCoop only supports the `authorization_code` grant at this time."
                    .to_string(),
            }),
        )
            .into_response();
    }
    let Some(auth_code) = form.code.clone() else {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidRequest,
                description: "`code` parameter missing.".to_string(),
            }),
        )
            .into_response();
    };
    let Some(auth_code_verifier) = form.code_verifier.clone() else {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidRequest,
                description: "`code_verifier` parameter missing.".to_string(),
            }),
        )
            .into_response();
    };
    let Ok(auth_code) = AuthCode::from_str(&auth_code) else {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidRequest,
                description: "`code` parameter malformed.".to_string(),
            }),
        )
            .into_response();
    };

    // Create an access token but don't actually issue it yet:
    // This lets us store the hash of the access token against the redemption of the auth code,
    // so double redemptions can invalidate the access token appropriately.
    let access_token = randgen.gen::<AccessToken>();
    let access_token_b64 = BASE64_URL_SAFE_NO_PAD.encode(access_token);
    let access_token_hash: AccessTokenHash = Blake2s256::digest(access_token).into();
    let refresh_token = randgen.gen::<RefreshToken>();
    let refresh_token_b64 = BASE64_URL_SAFE_NO_PAD.encode(refresh_token);
    let refresh_token_hash: RefreshTokenHash = Blake2s256::digest(refresh_token).into();

    // Redeem the auth code so we can check it and then maybe issue an access token.
    let binding = match code_store.redeem(&auth_code, access_token_hash, refresh_token_hash) {
        CodeRedemption::Invalid => {
            return (
                StatusCode::BAD_REQUEST,
                Json(TokenError {
                    code: TokenErrorCode::InvalidGrant,
                    description: "Auth code has expired or was not valid.".to_owned(),
                }),
            )
                .into_response();
        }
        CodeRedemption::Valid { binding } => binding,
        CodeRedemption::Conflicted {
            access_token_to_invalidate,
            refresh_token_to_invalidate,
        } => {
            // Invalidate the access token that was issued before
            if let Err(err) = store
                .txn(move |mut txn| {
                    Box::pin(async move {
                        txn.invalidate_access_token_by_hash(
                            &access_token_to_invalidate,
                            &refresh_token_to_invalidate,
                        )
                        .await
                        .context("failed to invalidate access token")
                    })
                })
                .await
            {
                error!("failed to handle Conflicted: {err:?}");
            }
            return (
                StatusCode::BAD_REQUEST,
                Json(TokenError {
                    code: TokenErrorCode::InvalidGrant,
                    description: "Auth code has been redeemed multiple times! This could mean something nasty.".to_owned(),
                }),
            )
                .into_response();
        }
    };

    // 1. Check the auth code is valid for this client
    if binding.client_id != basic_auth.username() {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidGrant,
                description: "Auth code is not valid for your client.".to_owned(),
            }),
        )
            .into_response();
    }
    if binding.redirect_uri != form.redirect_uri {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidGrant,
                description: "Auth code is not valid for your specific client redirect URI."
                    .to_owned(),
            }),
        )
            .into_response();
    }

    // 2. Check the code challenge
    let Some(computed_code_challenge) =
        compute_code_challenge(&binding.code_challenge_method, &auth_code_verifier)
    else {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidRequest,
                description: "Unable to compute code challenge here.".to_owned(),
            }),
        )
            .into_response();
    };

    if !bool::from(
        computed_code_challenge
            .as_bytes()
            .ct_eq(binding.code_challenge.as_bytes()),
    ) {
        return (
            StatusCode::BAD_REQUEST,
            Json(TokenError {
                code: TokenErrorCode::InvalidGrant,
                description: "Code challenge is invalid.".to_owned(),
            }),
        )
            .into_response();
    }

    let application_id = binding.client_id.clone();
    let user_id = binding.user_id;
    let user_login_session_id = binding.user_login_session_id;

    // Issue access token for a new session
    let clock2 = clock.clone();
    match store
        .txn(move |mut txn| {
            Box::pin(async move {
                let Some(session_id) = txn
                    .create_application_session(user_id, &application_id, user_login_session_id)
                    .await
                    .context("create_application_session")?
                else {
                    return Ok(Err((
                        StatusCode::BAD_REQUEST,
                        Json(TokenError {
                            code: TokenErrorCode::InvalidGrant,
                            description: "Auth code has expired or was not valid.".to_owned(),
                        }),
                    )
                        .into_response()));
                };
                txn.issue_access_token(
                    &access_token_hash,
                    session_id,
                    // TODO(expiry) Support custom expiry, not 100 years
                    clock2.now_utc() + Duration::days(365 * 100),
                )
                .await
                .context("issue_access_token")?;
                txn.issue_refresh_token(
                    &refresh_token_hash,
                    session_id,
                    // TODO(expiry) Support custom expiry, not 100 years
                    clock2.now_utc() + Duration::days(365 * 100),
                )
                .await
                .context("issue_refresh_token")?;

                Ok(Ok(session_id))
            })
        })
        .await
    {
        Ok(Ok(session_id)) => {
            debug!(
                "created application session {session_id:?} for {:?}",
                binding.client_id
            );
        }
        Ok(Err(web_err)) => {
            return web_err;
        }
        Err(err) => {
            error!("failed to issue access token: {err:?}");
            return (
            StatusCode::INTERNAL_SERVER_ERROR,
            "database error when issuing access token; ask the server administrator to check server logs"
        ).into_response();
        }
    }

    let now = clock.now_timestamp();
    // TODO(expiry) Support custom expiry times (not just 100 years)
    let exp = now + 100 * 365 * 86400;
    let sub = binding.user_id.hyphenated().to_string();
    let id_token = IdToken {
        iss: config.oidc.issuer.clone(),
        sub,
        aud: basic_auth.username().to_owned(),
        exp,
        iat: now,
        auth_time: 0, // TODO
        nonce: binding.nonce,
    };

    let id_token = match make_id_token(id_token, &secrets) {
        Ok(token) => token,
        Err(err) => {
            error!("Failed to make ID Token: {err:?}");
            return (StatusCode::INTERNAL_SERVER_ERROR, "failed to make ID Token").into_response();
        }
    };

    (
        StatusCode::OK,
        Json(TokenSuccess {
            access_token: access_token_b64,
            token_type: "Bearer".to_owned(),
            refresh_token: refresh_token_b64,
            expires_in: 86400 * 365, // TODO(expiry)
            // This assumes that we only support the OpenID scope at present.
            scope: "openid".to_owned(),
            id_token,
        }),
    )
        .into_response()
}

fn make_id_token(id_token: IdToken, secrets: &SecretConfig) -> eyre::Result<String> {
    let Ok(serde_json::Value::Object(map)) =
        serde_json::to_value(id_token).context("failed to serialise ID Token content")
    else {
        bail!("ID Token not a map");
    };

    let mut id_token_header = JwsHeader::new();
    id_token_header.set_token_type("JWT");
    // The key ID is currently hardcoded as `thekey`; this is the same
    // as on the JWKS endpoint.
    id_token_header.set_key_id("thekey");

    let id_token_payload =
        JwtPayload::from_map(map).context("failed to make JWT payload for ID token")?;

    let rs256_signer = Rs256
        .signer_from_jwk(&secrets.rsa_key_pair.to_jwk_key_pair())
        .context("failed to make RS256 signer")?;

    josekit::jwt::encode_with_signer(&id_token_payload, &id_token_header, &rs256_signer)
        .context("failed to sign ID Token")
}

/// Given a code challenge method and the code challenge verifier, compute
/// the 'code challenge'.
///
/// TODO(doc): provide spec reference for this
pub fn compute_code_challenge(method: &str, verifier: &str) -> Option<String> {
    match method {
        "plain" => Some(verifier.to_owned()),
        "S256" => {
            // code_challenge = BASE64URL-ENCODE(SHA256(ASCII(code_verifier)))
            let mut hasher = Sha256::new();
            hasher.update(verifier.as_bytes());
            let sha256_of_verifier = hasher.finalize();
            Some(BASE64_URL_SAFE_NO_PAD.encode(sha256_of_verifier))
        }
        _other => None,
    }
}

/// OpenID Connect ID Token
///
/// This is the payload of a JWT that is passed to the client/application.
///
/// TODO(doc): provide spec reference for this.
#[derive(Serialize)]
pub struct IdToken {
    /// REQUIRED. Issuer Identifier for the Issuer of the response.
    /// The iss value is a case sensitive URL using the https scheme that contains scheme, host, and optionally, port number and path components and no query or fragment components.
    pub iss: String,
    /// REQUIRED. Subject Identifier.
    /// A locally unique and never reassigned identifier within the Issuer for the End-User, which is intended to be consumed by the Client, e.g., 24400320 or AItOawmwtWwcT0k51BayewNvutrJUqsvl6qs7A4.
    /// It MUST NOT exceed 255 ASCII characters in length. The sub value is a case sensitive string.
    pub sub: String,
    /// REQUIRED. Audience(s) that this ID Token is intended for. It MUST contain the OAuth 2.0 client_id of the Relying Party as an audience value.
    /// It MAY also contain identifiers for other audiences. In the general case, the aud value is an array of case sensitive strings.
    /// In the common special case when there is one audience, the aud value MAY be a single case sensitive string.
    pub aud: String,
    /// REQUIRED. Expiration time on or after which the ID Token MUST NOT be accepted for processing.
    /// The processing of this parameter requires that the current date/time MUST be before the expiration date/time listed in the value.
    /// Implementers MAY provide for some small leeway, usually no more than a few minutes, to account for clock skew.
    /// Its value is a JSON number representing the number of seconds from 1970-01-01T0:0:0Z as measured in UTC until the date/time.
    /// See RFC 3339 [RFC3339] for details regarding date/times in general and UTC in particular.
    pub exp: i64,
    /// REQUIRED. Time at which the JWT was issued. Its value is a JSON number representing the number of seconds from 1970-01-01T0:0:0Z as measured in UTC until the date/time.
    pub iat: i64,
    /// Time when the End-User authentication occurred.
    /// Its value is a JSON number representing the number of seconds from 1970-01-01T0:0:0Z as measured in UTC until the date/time.
    /// When a max_age request is made or when auth_time is requested as an Essential Claim, then this Claim is REQUIRED; otherwise, its inclusion is OPTIONAL.
    /// (The auth_time Claim semantically corresponds to the OpenID 2.0 PAPE [OpenID.PAPE] auth_time response parameter.)
    pub auth_time: u64,
    /// String value used to associate a Client session with an ID Token, and to mitigate replay attacks.
    /// The value is passed through unmodified from the Authentication Request to the ID Token.
    /// If present in the ID Token, Clients MUST verify that the nonce Claim Value is equal to the value of the nonce parameter sent in the Authentication Request.
    /// If present in the Authentication Request, Authorization Servers MUST include a nonce Claim in the ID Token with the Claim Value being the nonce value sent in the Authentication Request.
    /// Authorization Servers SHOULD perform no other processing on nonce values used.
    /// The nonce value is a case sensitive string.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub nonce: Option<String>,
    // /// OPTIONAL. Authentication Context Class Reference. String specifying an Authentication Context Class Reference value that identifies the Authentication Context Class that the authentication performed satisfied. The value "0" indicates the End-User authentication did not meet the requirements of ISO/IEC 29115 [ISO29115] level 1. Authentication using a long-lived browser cookie, for instance, is one example where the use of "level 0" is appropriate. Authentications with level 0 SHOULD NOT be used to authorize access to any resource of any monetary value. (This corresponds to the OpenID 2.0 PAPE [OpenID.PAPE] nist_auth_level 0.) An absolute URI or an RFC 6711 [RFC6711] registered name SHOULD be used as the acr value; registered names MUST NOT be used with a different meaning than that which is registered. Parties using this claim will need to agree upon the meanings of the values used, which may be context-specific. The acr value is a case sensitive string.
    // pub acr: String,
    // /// OPTIONAL. Authentication Methods References. JSON array of strings that are identifiers for authentication methods used in the authentication. For instance, values might indicate that both password and OTP authentication methods were used. The definition of particular values to be used in the amr Claim is beyond the scope of this specification. Parties using this claim will need to agree upon the meanings of the values used, which may be context-specific. The amr value is an array of case sensitive strings.
    // pub amr: String,
    // /// OPTIONAL. Authorized party - the party to which the ID Token was issued. If present, it MUST contain the OAuth 2.0 Client ID of this party. This Claim is only needed when the ID Token has a single audience value and that audience is different than the authorized party. It MAY be included even when the authorized party is the same as the sole audience. The azp value is a case sensitive string containing a StringOrURI value.
    // pub azp: String,
}

#[derive(Serialize)]
struct TokenSuccess {
    /// The access token issued by the authorization server.
    access_token: String,
    /// literally just "Bearer" for us
    token_type: String,
    /// OPTIONAL. The refresh token, which can be used to obtain new access tokens based on the grant passed in the corresponding token request.
    refresh_token: String,
    /// RECOMMENDED. The lifetime in seconds of the access token. For example, the value 3600 denotes that the access token will expire in one hour from the time the response was generated.
    /// If omitted, the authorization server SHOULD provide the expiration time via other means or document the default value.
    expires_in: u64,
    /// RECOMMENDED, if identical to the scope requested by the client; otherwise, REQUIRED. The scope of the access token as described by Section 3.2.2.1.
    scope: String,
    /// OpenID Connect: ID Token value associated with the authenticated session.
    id_token: String,
}

#[derive(Serialize)]
struct TokenError {
    #[serde(rename = "error")]
    code: TokenErrorCode,

    #[serde(rename = "error_description")]
    description: String,
    // OPTIONAL error_uri
}

#[derive(Serialize)]
enum TokenErrorCode {
    #[serde(rename = "invalid_request")]
    InvalidRequest,
    #[serde{rename = "invalid_client"}]
    InvalidClient,
    #[serde{rename = "invalid_grant"}]
    InvalidGrant,
    #[allow(dead_code)]
    #[serde{rename = "unauthorized_client"}]
    UnauthorisedClient,
    #[serde{rename = "unsupported_grant_type"}]
    UnsupportedGrantType,
    #[allow(dead_code)]
    #[serde{rename = "invalid_scope"}]
    InvalidScope,
}
