//! Implements a simple anti-XSRF measure for when we have no login session active (see LoginSession for one to use if we DO have a login session).
//1
//! Tnis implements a simple 'Double Submit Cookie' pattern.
//!
//! Because we use a `__Host-` prefixed cookie, this is not vulnerable to any cookie fixation attacks in modern browsers.
//!
//! Even on older browsers, that type of attack is rare, so this 'naïve' scheme is fine for the purpose (rather than having to do anything complicated).

use base64::{prelude::BASE64_URL_SAFE_NO_PAD, Engine};
use rand::Rng;
use subtle::ConstantTimeEq;
use time::Duration;
use tower_cookies::{Cookie, Cookies};

use crate::utils::RandGen;

/// Name of the cookie for the sessionless cross-site request forgery prevention cookie.
pub const COOKIE_NAME: &str = "__Host-SessionlessXsrf";

/// Gets the Sessionless XSRF token to put into a form request
pub fn get_token(cookies: &Cookies, randgen: &mut RandGen) -> String {
    if let Some(xsrf_cookie) = cookies.get(COOKIE_NAME) {
        xsrf_cookie.value().to_owned()
    } else {
        let new_token = randgen.gen::<[u8; 16]>();
        let new_token_b64 = BASE64_URL_SAFE_NO_PAD.encode(new_token);
        cookies.add(
            Cookie::build(COOKIE_NAME, new_token_b64.clone())
                .path("/")
                .http_only(true)
                .secure(true)
                .same_site(tower_cookies::cookie::SameSite::Strict)
                .max_age(Duration::days(500))
                .finish(),
        );
        new_token_b64
    }
}

/// Checks a Sessionless XSRF token obtained from a form request
pub fn check_token(cookies: &Cookies, token: &str) -> bool {
    let Some(xsrf_token) = cookies.get(COOKIE_NAME) else {
        return false;
    };

    bool::from(xsrf_token.value().as_bytes().ct_eq(token.as_bytes()))
}
